##
##

import logging
import attr
import os
import json
import argparse
import yaml
import couchformation.constants as C
from typing import Optional, List, Tuple, Union, Any
from enum import Enum
from couchformation.exception import FatalError
from couchformation.config import BaseConfig, NodeConfig, Parameters, AuthMode, get_project_dir, get_state_file, State
from couchformation.util import FileManager, dict_merge, dict_merge_not_none
from couchformation.kvdb import KeyValueStore
from couchformation.util import PasswordUtility, UUIDGen
from couchformation.certificates import CertMgr

DEPLOYMENT = "deployment.db"
logger = logging.getLogger('couchformation.deployment')
logger.addHandler(logging.NullHandler())


class DeploymentError(FatalError):
    pass


class ServiceModel(str, Enum):
    cbs = 'cbs'
    sgw = 'sgw'
    sdk = 'sdk'
    generic = 'generic'


@attr.s
class Service:
    cloud: Optional[str] = attr.ib(default="aws")
    name: Optional[str] = attr.ib(default="nodes")
    model: Optional[str] = attr.ib(default=ServiceModel.generic)
    region: Optional[str] = attr.ib(default=None)
    os_id: Optional[str] = attr.ib(default=None)
    os_version: Optional[str] = attr.ib(default=None)
    auth_mode: Optional[str] = attr.ib(default="default")
    profile: Optional[str] = attr.ib(default='default')
    connect_svc: Optional[str] = attr.ib(default=None)
    connect_ip: Optional[str] = attr.ib(default=None)
    config: Optional[List[NodeConfig]] = attr.ib(default=[])

    def from_dict(self, options: dict):
        for attribute in self.__annotations__:
            if attribute == 'config':
                continue
            if options.get(attribute):
                setattr(self, attribute, options.get(attribute))

    @property
    def auth(self) -> AuthMode:
        return AuthMode[self.auth_mode]


@attr.s
class BuildParameter:
    name: Optional[str] = attr.ib()
    type: Optional[str] = attr.ib(default="string")
    allowed_values: Optional[List[str]] = attr.ib(default=[])
    required_values: Optional[List[str]] = attr.ib(default=[])


@attr.s
class Build:
    name: Optional[str] = attr.ib()
    parameters: Optional[List[BuildParameter]] = attr.ib(default=[])
    supports: Optional[List[str]] = attr.ib(default=[])


@attr.s
class BuildList:
    build_list: List[Build] = attr.ib(default=[])

    def add(self, build: Build):
        self.build_list.append(build)

    def get(self, name: str) -> Union[Build, None]:
        for build in self.build_list:
            if build.name == name:
                return build
        return None


class MetadataManager(object):

    def __init__(self, project: str):
        self.project = project
        self.project_dir = get_project_dir(project)
        self.metadata = os.path.join(self.project_dir, C.METADATA)
        self.network = os.path.join(self.project_dir, C.NETWORK)
        self.credentials = os.path.join(self.project_dir, C.CREDENTIALS)

    @property
    def project_uid(self):
        meta = KeyValueStore(self.metadata)
        meta.document('config')
        return meta.get('project_uid')

    @property
    def exists(self):
        return os.path.exists(self.metadata)

    def list_services(self):
        if not self.exists:
            return ()
        meta = KeyValueStore(self.metadata)
        meta.document('resources')
        for resource in meta.keys():
            if not resource:
                continue
            yield resource, meta[resource]

    def list_cloud_net_regions(self, cloud: str):
        db = KeyValueStore(self.network)
        doc_list = db.doc_id_startswith(f"network:{cloud}")
        return [doc.split(':')[2] for doc in doc_list if len(doc.split(':')) == 3]

    def get_service_groups(self, service: str) -> List[KeyValueStore]:
        filename = os.path.join(self.project_dir, f"{service}.db")
        db = KeyValueStore(filename)
        doc_list = db.doc_id_startswith(service)
        return [KeyValueStore(filename, doc) for doc in doc_list]

    def get_network_state(self, cloud: str, region: str):
        filename = get_state_file(self.project, f"network-{region}")
        if not os.path.exists(filename):
            return {}
        document = f"network:{cloud}"
        state = KeyValueStore(filename, document)
        return state.as_dict

    def get_network_params(self, cloud: str, region: str):
        document = f"network:{cloud}:{region}"
        parameters = KeyValueStore(self.network, document)
        return parameters.as_dict

    def get_service_state(self, service: str, number: int):
        filename = get_state_file(self.project, service)
        if not os.path.exists(filename):
            return {}
        node_name = f"{service}-node-{number:02d}"
        state = KeyValueStore(filename, node_name)
        return state.as_dict

    def get_project_ca(self):
        document = f"credentials:{self.project}"
        credentials = KeyValueStore(self.credentials)
        credentials.document(document)
        return credentials.get('ca_cert')

    def copy_project(self, target: str):
        target_dir = get_project_dir(target)

        if not self.exists:
            raise DeploymentError(f"Project {self.project} does not exist")

        target_metadata = os.path.join(target_dir, C.METADATA)
        target_network = os.path.join(target_dir, C.NETWORK)
        try:
            FileManager().make_dir(target_dir)
            FileManager().copy_file(self.metadata, target_metadata)
            FileManager().copy_file(self.network, target_network)
            db = KeyValueStore(target_network)
            doc_list = db.doc_id_startswith('network')
            for doc in doc_list:
                net = KeyValueStore(db.file_name, doc)
                net['project'] = target
        except Exception as err:
            raise DeploymentError(f"can not create target project dir: {err}")

        for service, cloud in self.list_services():
            filename = f"{service}.db"
            source_filename = os.path.join(self.project_dir, filename)
            target_filename = os.path.join(target_dir, filename)
            FileManager().copy_file(source_filename, target_filename)
            db = KeyValueStore(target_filename)
            doc_list = db.doc_id_startswith(service)
            for doc in doc_list:
                resource = KeyValueStore(db.file_name, doc)
                resource['project'] = target

    @staticmethod
    def service_state(state_data: dict):
        if state_data.get('state', 0) == State.DEPLOYED.value:
            return f" : {C.GREEN_COLOR}Deployed{C.SCREEN_RESET}"
        elif state_data.get('state', 0) == State.DEPLOYING.value:
            return f" : {C.BLUE_COLOR}Deploying{C.SCREEN_RESET}"
        elif state_data.get('state', 0) == State.FAILED.value:
            return f" : {C.RED_COLOR}Failed{C.SCREEN_RESET}"
        else:
            return ""

    def network_cidr(self, cloud: str, region: str, net_data: dict):
        if not net_data:
            net_config = self.get_network_params(cloud, region)
            cidr = net_config.get('cidr')
        else:
            if net_data.get('vpc_cidr'):
                cidr = net_data.get('vpc_cidr')
            elif net_data.get('network_cidr'):
                cidr = net_data.get('network_cidr')
            else:
                cidr = None
        return f"CIDR: {cidr}" if cidr else ""

    def print_services(self):
        log = logging.getLogger('minimum_output')
        cloud_map = {}
        log.info(f"\n[[{self.project}]]")
        for service, cloud in self.list_services():
            cloud_map.setdefault(cloud, []).append(service)
        for cloud in cloud_map.keys():
            log.info(f"Cloud: {cloud}")
            log.info("[network]")
            for net_region in self.list_cloud_net_regions(cloud):
                if net_region is None or net_region == "None":
                    continue
                net_data = self.get_network_state(cloud, net_region)
                net_status = self.service_state(net_data)
                net_cidr = self.network_cidr(cloud, net_region, net_data)
                log.info(f"+- [{net_region}]")
                if net_cidr:
                    log.info(f"| +- {net_cidr}{net_status}")
            log.info("[services]")
            for service in cloud_map[cloud]:
                log.info(f"+- [{service}]")
                for n, group in enumerate(self.get_service_groups(service)):
                    space = " "
                    state_data = self.get_service_state(service, n + 1)
                    status = self.service_state(state_data)
                    build = group['build'] if 'build' in group and group['build'] is not None else ''
                    region = group['region'] if 'region' in group and group['region'] is not None else ''
                    os_id = group['os_id'] if 'os_id' in group and group['os_id'] is not None else ''
                    machine_type = group['machine_type'] if 'machine_type' in group and group['machine_type'] is not None else ''
                    quantity = group['quantity'] if 'quantity' in group and group['quantity'] is not None else 1
                    log.info(f"| +- [{n+1}] ({build}) {quantity}x {os_id}{space * bool(os_id)}{machine_type} {region}{status}")

    def print_cli(self, options: argparse.Namespace):
        log = logging.getLogger('minimum_output')
        cloud_map = {}
        for service, cloud in self.list_services():
            cloud_map.setdefault(cloud, []).append(service)
        for cloud in cloud_map.keys():
            for service in cloud_map[cloud]:
                log.info(f"[{service}]")
                for n, group in enumerate(self.get_service_groups(service)):
                    command = "create" if n == 0 else "add"
                    out_line = f"cloudmgr {command} "
                    for attribute in vars(options):
                        if attribute == 'command' or attribute == 'group' or attribute == 'provisioner':
                            continue
                        if group.get(attribute):
                            out_line += f"--{attribute} {group[attribute]} "
                    for parameter in group:
                        if parameter not in vars(options) and group[parameter] is not None:
                            out_line += f"--{parameter} {group[parameter]} "
                    log.info(out_line)

    def print_project(self, options: argparse.Namespace, name: str = None):
        log = logging.getLogger('minimum_output')
        cloud_map = {}
        for service, cloud in self.list_services():
            cloud_map.setdefault(cloud, []).append(service)
        for cloud in cloud_map.keys():
            for service in cloud_map[cloud]:
                if name and service != name:
                    continue
                log.info(f"\n<Service: {service}>")
                for n, group in enumerate(self.get_service_groups(service)):
                    if n == 0:
                        log.info(f"Cloud: {group.get('cloud')}")
                        log.info(f"Build: {group.get('build')}")
                    log.info(f"[{service}] Group: {n+1}")
                    for key, value in group.items():
                        if key not in vars(options):
                            continue
                        if value is not None:
                            log.info(f"{key:<14}: {value}")

    def edit_service(self, name: str, group: int, options: argparse.Namespace):
        if len([attribute for attribute in vars(options) if getattr(options, attribute) is not None]) == 0:
            raise DeploymentError(f"No parameters to edit for service {name}")
        cloud_map = {}
        for service, cloud in self.list_services():
            cloud_map.setdefault(cloud, []).append(service)
        for cloud in cloud_map.keys():
            for service in cloud_map[cloud]:
                if service != name:
                    continue
                filename = f"{service}.db"
                service_filename = os.path.join(self.project_dir, filename)
                db = KeyValueStore(service_filename)
                document = f"{service}:{group:04d}"
                db.document(document)
                for attribute in vars(options):
                    if getattr(options, attribute) is None:
                        continue
                    logger.info(f"Setting service {service} parameter {attribute} to \"{getattr(options, attribute)}\"")
                    db[attribute] = getattr(options, attribute)
                region = db.get('region', 'local')
                document = f"network:{cloud}:{region}"
                net = KeyValueStore(self.network)
                net.document(document)
                for attribute in vars(options):
                    if getattr(options, attribute) is None:
                        continue
                    logger.debug(f"Setting network db parameter {attribute} to \"{getattr(options, attribute)}\"")
                    net[attribute] = getattr(options, attribute)
        self.print_project(options, name)


class BuildManager(object):

    def __init__(self, options: argparse.Namespace, parameters: List[str]):
        self.cfg_file = C.BUILD_PROFILES
        self.options = options
        self.parameters = parameters
        self.create_mode = self.options.command == "create"
        self.build_list = BuildList()
        self.load_config()

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for build, settings in yaml.safe_load(f).items():
                    p_list = []
                    for parameter in settings.get('parameters', []):
                        for key, value in parameter.items():
                            p_list.append(BuildParameter(key, value.get('type'), value.get('allowed_values'), value.get('required_values')))
                    build = Build(build, p_list, settings.get('supports', []))
                    self.build_list.add(build)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open build config file {self.cfg_file}: {err}")

    def validate(self):
        build = self.build_list.get(self.options.build)
        if not build:
            return None
        self.validate_base()
        self.validate_command(build)
        self.validate_parameters(build)

    def validate_base(self):
        project = self.options.project
        name = self.options.name
        services = [service for service, cloud in MetadataManager(project).list_services()]
        if name in services and self.create_mode:
            logger.warning(f"Overwriting previously configured service {name}")

    def validate_command(self, build: Build):
        command = self.options.command
        if command not in build.supports:
            raise DeploymentError(f"Command {command} not allowed with build type {build.name}")

    def validate_parameters(self, build: Build):
        for parameter in build.parameters:
            if not self.parameter_check(parameter):
                raise DeploymentError(f"--{parameter.name} value is not valid")

    def parameter_check(self, parameter: BuildParameter):
        parser = argparse.ArgumentParser(add_help=False)
        if parameter.type == "boolean":
            parser.add_argument(f"--{parameter.name}", action='store_true')
        else:
            parser.add_argument(f"--{parameter.name}", action='store')
        options, undefined = parser.parse_known_args(self.parameters)
        result = getattr(options, parameter.name)
        if result:
            return self.value_check(parameter.type, result, parameter.allowed_values, parameter.required_values)
        return True

    def value_check(self, v_type: str, v_value: Any, v_allowed: List[str], v_required: List[str]):
        if v_type == "boolean":
            return isinstance(v_value, bool)
        elif v_type == "integer":
            try:
                int(v_value)
            except ValueError:
                return False
        elif v_type == "float":
            try:
                float(v_value)
            except ValueError:
                return False
        elif v_type == "string":
            pass
        elif v_type == "csv":
            if not isinstance(v_value, str) or not all(val in v_allowed for val in v_value.split(',')):
                logger.warning(f"Invalid comma separated list: allowed values: {','.join(v_allowed)}")
                return False
            if self.create_mode and not any(val in v_required for val in v_value.split(',')):
                logger.warning(f"Invalid comma separated list: required values: {','.join(v_required)}")
                return False
        elif v_type == "path":
            if not os.path.exists(v_value):
                logger.warning(f"Path does not exist: {v_value}")
                return False
        return True


class NodeGroup(object):

    def __init__(self, options: argparse.Namespace):
        self.options = options
        self.project = self.options.project
        self.name = self.options.name
        self.cloud = self.options.cloud
        self.project_dir = get_project_dir(self.project)

        try:
            FileManager().make_dir(self.project_dir)
        except Exception as err:
            raise DeploymentError(f"can not create project dir: {err}")

        filename = os.path.join(self.project_dir, f"{self.name}.db")
        network = os.path.join(self.project_dir, C.NETWORK)
        metadata = os.path.join(self.project_dir, C.METADATA)
        credentials = os.path.join(self.project_dir, C.CREDENTIALS)
        self.db = KeyValueStore(filename)
        self.net = KeyValueStore(network)
        self.meta = KeyValueStore(metadata)
        self.credentials = KeyValueStore(credentials)

        self.meta.document('config')
        if not self.meta['project_uid']:
            self.project_uid = UUIDGen().get_project_uid(self.project)
            self.meta['project_uid'] = self.project_uid
        else:
            self.project_uid = self.meta['project_uid']

    def create_network(self, parameters: argparse.Namespace, region, group=1):
        document = f"network:{self.cloud}:{region}"
        self.net.document(document)

        opt_dict = vars(self.options)
        parm_dict = vars(parameters)
        opt_merge = dict_merge(opt_dict, parm_dict)
        combined = dict_merge_not_none(self.net.as_dict, opt_merge)

        if group == 1:
            self.meta.document('network')
            self.net.update(combined)
            self.meta[self.cloud] = True

    def create_credentials(self):
        document = f"credentials:{self.project}"

        self.credentials.document(document)
        if not self.credentials.get('password'):
            password = PasswordUtility().generate(16)
            self.credentials['password'] = password

        return self.credentials.get('password')

    def create_ca(self):
        document = f"credentials:{self.project}"

        self.credentials.document(document)

        if not self.credentials.get('private_key'):
            private_key = CertMgr().private_key()
            self.credentials['private_key'] = private_key
        else:
            private_key = self.credentials.get('private_key')

        if not self.credentials.get('ca_cert'):
            ca_cert = CertMgr().certificate_ca(private_key)
            self.credentials['ca_cert'] = ca_cert

        return self.credentials.get('private_key'), self.credentials.get('ca_cert')

    def remove_credentials(self):
        document = f"credentials:{self.project}"
        self.credentials.remove(document)

    def create_node_group(self, parameters: argparse.Namespace, group=1):
        document = f"{self.name}:{group:04d}"
        region = parameters.region if 'region' in parameters else "local"

        if not self.name:
            raise DeploymentError(f"name is required")

        opt_dict = vars(self.options)
        parm_dict = vars(parameters)
        combined = dict_merge(opt_dict, parm_dict)

        if group == 1:
            self.db.clean()
        self.db.document(document)
        self.meta.document('resources')
        self.db.update(combined)
        self.meta[self.name] = self.cloud

        self.create_network(parameters, region, group)

    def get_credentials(self):
        document = f"credentials:{self.project}"
        self.credentials.document(document)
        return self.credentials.get('password')

    def add_to_node_group(self, parameters: argparse.Namespace):
        count = len(self.db.doc_id_startswith(self.name))
        if count == 0:
            raise ValueError(f"attempting to add to empty node group")
        self.create_node_group(parameters, count + 1)

    def get_node_groups(self) -> List[KeyValueStore]:
        self.meta.document('resources')
        for resource in self.meta.keys():
            if not resource:
                continue
            filename = os.path.join(get_project_dir(self.project), f"{resource}.db")
            db = KeyValueStore(filename)
            doc_list = db.doc_id_startswith(resource)
            yield [KeyValueStore(filename, doc) for doc in doc_list]

    def remove_node_groups(self, name=None):
        self.meta.document('resources')
        for resource in self.meta.keys():
            if name and resource != name:
                continue
            if not resource:
                continue
            filename = os.path.join(get_project_dir(self.project), f"{resource}.db")
            db = KeyValueStore(filename)
            doc_list = db.doc_id_startswith(resource)
            for doc in doc_list:
                db.remove(doc)
            del self.meta[resource]

    def clean_node_groups(self, name=None):
        self.meta.document('resources')
        for resource in self.meta.keys():
            if name and resource != name:
                continue
            filename = os.path.join(get_project_dir(self.project), f"{resource}.db")
            logger.info(f"Removing {resource} database")
            os.remove(filename)

    def clean_base(self):
        logger.info("Removing project core databases")
        network = os.path.join(self.project_dir, C.NETWORK)
        os.remove(network)
        metadata = os.path.join(self.project_dir, C.METADATA)
        os.remove(metadata)

    def get_networks(self) -> List[KeyValueStore]:
        doc_list = self.net.doc_id_startswith('network')
        return [KeyValueStore(self.net.file_name, doc) for doc in doc_list]

    def get_network(self, cloud, region):
        doc_list = self.net.doc_id_startswith('network')
        return next((KeyValueStore(self.net.file_name, doc) for doc in doc_list if doc.endswith(f"{cloud}:{region}")), None)


class Deployment(object):

    def __init__(self, parameters: Parameters):
        self.parameters = parameters
        self.core, self.deploy_set = self.from_file()

        try:
            FileManager().make_dir(self.parameters.project_dir)
        except Exception as err:
            raise DeploymentError(f"can not create working dir: {err}")

    @property
    def services(self) -> Tuple[str, BaseConfig, Service]:
        for name, service in self.deploy_set.items():
            yield name, self.core, service

    def store_config(self, overwrite: bool = False):
        service = self.parameters.name
        if not self.deploy_set.get(service) or overwrite:
            new_svc = Service(config=[])
            new_svc.from_dict(self.parameters.as_dict)
            self.deploy_set.update({service: new_svc})
        config = NodeConfig().create(self.parameters.as_dict)
        self.deploy_set.get(service).config.append(config)

    def from_file(self):
        deploy_set = {}
        core = BaseConfig()
        config_data = self.read_file("deployment.cfg")

        if config_data.get('core'):
            core.from_dict(config_data.get('core'))

        core.from_dict(self.parameters.as_dict)

        for name, service in config_data.items():
            if name == 'core':
                continue
            new_svc = Service(config=[])
            new_svc.from_dict(service)
            for saved_config in service.get('config'):
                config = NodeConfig().create(saved_config)
                new_svc.config.append(config)
            deploy_set.update({name: new_svc})

        return core, deploy_set

    def to_file(self):
        config_data = {
            'core': self.core.as_dict
        }
        for name, service in self.deploy_set.items():
            # noinspection PyTypeChecker
            config_data.update({name: attr.asdict(service)})
        self.write_file(config_data, "deployment.cfg")

    def read_file(self, name: str):
        cfg_file = os.path.join(self.parameters.project_dir, name)
        try:
            with open(cfg_file, 'r') as cfg_file_h:
                data = json.load(cfg_file_h)
                return data
        except FileNotFoundError:
            return {}
        except Exception as err:
            raise DeploymentError(f"can not read from config file {cfg_file}: {err}")

    def write_file(self, data: dict, name: str):
        cfg_file = os.path.join(self.parameters.project_dir, name)
        try:
            with open(cfg_file, 'w') as cfg_file_h:
                json.dump(data, cfg_file_h, indent=2)
                cfg_file_h.write('\n')
        except Exception as err:
            raise DeploymentError(f"can not write to config file {cfg_file}: {err}")



##
##

import re
import logging
from typing import List, Union
import googleapiclient.errors
from couchformation.gcp.driver.base import CloudBase, GCPDriverError, EmptyResultSet

logger = logging.getLogger('couchformation.gcp.driver.firewall')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class Firewall(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self) -> List[dict]:
        firewall_list = []

        try:
            request = self.gcp_client.firewalls().list(project=self.gcp_project)
            while request is not None:
                response = request.execute()

                for firewall in response['items']:
                    firewall_list.append(firewall)
                request = self.gcp_client.firewalls().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverError(f"error listing firewall rules: {err}")

        if len(firewall_list) == 0:
            raise EmptyResultSet(f"no firewalls found")
        else:
            return firewall_list

    def search(self, pattern: str) -> List[dict]:
        firewall_list = []
        for entry in self.list():
            if re.search(pattern, entry['name']):
                firewall_list.append(entry)
        return firewall_list

    def create_ingress(self, name: str, network: str, cidr: str, protocol: str = "tcp", ports: Union[List[str], None] = None, udp_ports: Union[List[str], None] = None) -> str:
        operation = {}
        firewall_body = {
            "sourceRanges": [
                cidr,
            ],
            "description": "Couch Formation generated firewall rule",
            "allowed": [
                {
                    "IPProtocol": protocol,
                },
            ],
            "network": f"global/networks/{network}",
            "name": name,
        }
        if ports:
            firewall_body['allowed'][0]['ports'] = []
            firewall_body['allowed'][0]['ports'].extend(ports)
        if udp_ports:
            firewall_body['allowed'].append(dict(
                IPProtocol="udp",
                ports=udp_ports
            ))
        try:
            request = self.gcp_client.firewalls().insert(project=self.gcp_project, body=firewall_body)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create firewall rule: {err}")
        except Exception as err:
            raise GCPDriverError(f"error creating firewall rule: {err}")

        return operation.get('targetLink')

    def delete(self, firewall: str) -> None:
        try:
            request = self.gcp_client.firewalls().delete(project=self.gcp_project, firewall=firewall)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not delete firewall rule: {err}")
        except Exception as err:
            raise GCPDriverError(f"error deleting firewall rule: {err}")

    def details(self, firewall: str) -> Union[dict, None]:
        try:
            request = self.gcp_client.firewalls().get(project=self.gcp_project, firewall=firewall)
            result = request.execute()
            return result
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find firewall entry: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting firewall rule: {err}")



##
##

import logging
from typing import Union
from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt.compute.models import DiskCreateOption
from couchformation.azure.driver.base import CloudBase, AzureDriverError

logger = logging.getLogger('couchformation.azure.driver.disk')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Disk(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def create(self, resource_group: str, location: str, zone: str, size_value: Union[int, str], name: str, ultra: bool = False):
        size = int(size_value)
        disk_perf = self.disk_size_to_tier(size)
        parameters = {
            'location': location,
            'sku': {},
            'zones': [zone],
            'disk_size_gb': int(disk_perf['disk_size']),
            'creation_data': {
                'create_option': DiskCreateOption.empty
            }
        }
        if ultra:
            parameters['sku']['name'] = 'UltraSSD_LRS'
            parameters['disk_iops_read_write'] = int(disk_perf['disk_iops'])
        else:
            parameters['sku']['name'] = 'Premium_LRS'
            parameters['tier'] = disk_perf['disk_tier']
        try:
            request = self.compute_client.disks.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating disk: {err}")

    def details(self, name: str, resource_group: str) -> Union[dict, None]:
        try:
            disk = self.compute_client.disks.get(resource_group, name)
            disk_info = {'name': disk.name,
                         'id': disk.id,
                         'zones': disk.zones,
                         'disk_size_gb': disk.disk_size_gb,
                         'sku': disk.sku.__dict__}
            return disk_info
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting disk {name}: {err}")

    def delete(self, name: str, resource_group: str) -> None:
        try:
            request = self.compute_client.disks.begin_delete(resource_group, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting disk: {err}")



##
##

import logging
import re
from datetime import datetime
from typing import Union, List
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSEbsDisk, AWSTagStruct, EbsVolume, AWSTag, AWSImageOwners
import couchformation.constants as C

logger = logging.getLogger('couchformation.aws.driver.image')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class Image(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, filter_keys_exist: Union[List[str], None] = None, is_public: bool = False, owner_id: str = None, name: str = None) -> List[dict]:
        image_list = []
        if owner_id:
            owner_filter = [owner_id]
        else:
            owner_filter = []
        if is_public:
            ami_filter = [
                {
                    'Name': 'architecture',
                    'Values': [
                        "x86_64",
                        "arm64",
                        "arm64_mac"
                    ]
                },
                {
                    'Name': 'root-device-type',
                    'Values': [
                        "ebs",
                    ]
                }
            ]
            if name:
                ami_filter.append(
                    {
                        'Name': 'name',
                        'Values': [
                            name
                        ]
                    }
                )
        else:
            ami_filter = [
                {
                    'Name': 'is-public',
                    'Values': [
                        'false',
                    ]
                }
            ]

        logger.debug(f"Searching images by owner {owner_id} and name {name}")

        try:
            images = self.ec2_client.describe_images(Filters=ami_filter, Owners=owner_filter, IncludeDeprecated=True)
        except Exception as err:
            raise AWSDriverError(f"error getting AMIs: {err}")

        for image in images['Images']:
            image_block = {'name': image['ImageId'],
                           'description': image['Name'],
                           'date': image['CreationDate'],
                           'details': image['PlatformDetails'],
                           'root_disk': image.get('BlockDeviceMappings')[0].get('DeviceName'),
                           'arch': image['Architecture']}
            if is_public:
                image_block.update({'owner': image['OwnerId']})

            if filter_keys_exist:
                if not all(key in image_block for key in filter_keys_exist):
                    continue
            image_list.append(image_block)

        if len(image_list) == 0:
            raise EmptyResultSet(f"no AMIs found")

        return image_list

    def list_standard(self, architecture: str = 'x86_64', os_id: str = None, os_version: str = None, feature: str = None):
        result_list = []
        for image_type in AWSImageOwners.image_owner_list:
            if os_id and (image_type['os_id'] != os_id or image_type['feature'] != feature):
                continue
            image_list = self.list(is_public=True, owner_id=image_type['owner_id'], name=image_type['pattern'])
            for version in C.OS_VERSION_LIST[image_type['os_id']]:
                logger.debug(f"Checking image {version}")
                if os_version and version != os_version:
                    continue
                filtered_images = []
                for image in image_list:
                    if C.DEBUG_VERBOSE_LEVEL >= 3:
                        logger.debug(f"Found image {image}")
                    if image['arch'] != architecture:
                        continue
                    match = re.search(image_type['version'], image['description'])
                    if match and match.group(1) == version:
                        filtered_images.append(image)
                if len(filtered_images) > 0:
                    filtered_images.sort(key=lambda i: datetime.strptime(i['date'], '%Y-%m-%dT%H:%M:%S.%fZ'))
                    result_image = filtered_images[-1]
                    result_image.update(dict(
                        os_id=image_type['os_id'],
                        os_version=version,
                        os_user=image_type['user']
                    ))
                    result_list.append(result_image)
        if len(result_list) > 0:
            logger.debug(f"Selected image -> {result_list[-1]}")
        return result_list[-1]

    @staticmethod
    def image_user(os_id: str):
        return next((image_type['user'] for image_type in AWSImageOwners.image_owner_list if image_type['os_id'] == os_id), None)

    def details(self, ami_id: str) -> dict:
        ami_filter = {
            'Name': 'image-id',
            'Values': [
                ami_id,
            ]
        }
        try:
            result = self.ec2_client.describe_images(Filters=[ami_filter])
        except Exception as err:
            raise AWSDriverError(f"error getting AMI details: {err}")

        return result['Images'][0]

    def create(self, name: str, instance: str, description=None, root_type="gp3", root_size=100, root_iops=3000) -> str:
        try:
            instance_details = self.instance_details(instance)
        except Exception as err:
            raise AWSDriverError(f"error getting instance {instance} details: {err}")
        if 'BlockDeviceMappings' not in instance_details:
            raise AWSDriverError(f"can not get details for instance {instance}")

        root_dev = instance_details['BlockDeviceMappings'][0]['DeviceName']
        root_disk = [AWSEbsDisk.build(root_dev, EbsVolume(root_type, root_size, root_iops)).as_dict]
        ami_tag = [AWSTagStruct.build("image").add(AWSTag("Name", name)).as_dict]

        if not description:
            description = "couch-formation-image"

        try:
            result = self.ec2_client.create_image(BlockDeviceMappings=root_disk,
                                                  Description=description,
                                                  InstanceId=instance,
                                                  Name=name,
                                                  TagSpecifications=ami_tag)
        except Exception as err:
            raise AWSDriverError(f"error creating AMI: {err}")

        ami_id = result['ImageId']
        waiter = self.ec2_client.get_waiter('image_available')
        waiter.wait(ImageIds=[ami_id])

        return ami_id

    def delete(self, ami: str) -> None:
        try:
            self.ec2_client.deregister_image(ImageId=ami)
        except Exception as err:
            raise AWSDriverError(f"error deleting AMI: {err}")

    def instance_details(self, instance_id: str) -> dict:
        try:
            result = self.ec2_client.describe_instances(InstanceIds=[instance_id])
        except Exception as err:
            raise AWSDriverError(f"error getting instance details: {err}")

        return result['Reservations'][0]['Instances'][0]



##
##

import logging
import boto3
import botocore.exceptions
import botocore.session
from botocore.config import Config
import os
import webbrowser
import time
import configparser
from pathlib import Path
from datetime import datetime, timezone
from couchformation.exception import FatalError, NonFatalError
from couchformation.config import AuthMode

logger = logging.getLogger('couchformation.aws.driver.base')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class AWSDriverError(FatalError):
    pass


class EmptyResultSet(NonFatalError):
    pass


class CloudBase(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.config_directory = os.path.join(Path.home(), '.aws')
        self.config_file = os.path.join(self.config_directory, 'config')
        self.credential_file = os.path.join(self.config_directory, 'credentials')
        self.config_data = configparser.ConfigParser()
        self.credential_data = configparser.ConfigParser()
        self.profile = parameters.get('profile') if parameters.get('profile') else 'default'
        self.sso_session = None
        self.sso_account_id = None
        self.sso_role_name = None
        self.sso_start_url = None
        self.sso_region = None
        self.sso_registration_scopes = None
        self.profile_region = None
        self.aws_region = None
        self.zone_list = []
        self.access_key = None
        self.secret_key = None
        self.token = None
        self.token_expiration = None
        self.timeouts = Config(
            connect_timeout=1,
            read_timeout=1,
            retries={'max_attempts': 2}
        )

        self.read_config()

        if not parameters.get('auth_mode') or AuthMode[parameters.get('auth_mode')] == AuthMode.default:
            self.default_auth()
        else:
            self.sso_auth()

        if parameters.get('region'):
            os.environ['AWS_DEFAULT_REGION'] = parameters.get('region')
        elif self.profile_region:
            os.environ['AWS_DEFAULT_REGION'] = self.profile_region

        if self.access_key:
            os.environ['AWS_ACCESS_KEY_ID'] = self.access_key
        if self.secret_key:
            os.environ['AWS_SECRET_ACCESS_KEY'] = self.secret_key
        if self.token:
            os.environ['AWS_SESSION_TOKEN'] = self.token

        try:
            self.ec2_client = boto3.client('ec2', region_name=self.aws_region)
            self.dns_client = boto3.client('route53')
            self.sts_client = boto3.client('sts')
            # self.cost_client = boto3.client('pricing', region_name='us-east-1')
        except Exception as err:
            raise AWSDriverError(f"can not initialize AWS driver: {err}")

    @property
    def account_id(self):
        return self.sts_client.get_caller_identity()["Account"]

    def test_session(self):
        try:
            client = boto3.client('s3', region_name=self.aws_region)
            client.list_buckets()
        except Exception as err:
            raise AWSDriverError(f"not authorized: {err}")

    def read_config(self):
        if os.path.exists(self.config_file):
            try:
                self.config_data.read(self.config_file)
            except Exception as err:
                raise AWSDriverError(f"can not read config file {self.config_file}: {err}")

        if os.path.exists(self.credential_file):
            try:
                self.credential_data.read(self.credential_file)
            except Exception as err:
                raise AWSDriverError(f"can not read config file {self.credential_file}: {err}")

    def read_sso_config(self):
        for section, contents in self.config_data.items():
            if section.startswith('profile'):
                profile_name = section.split()[1]
                if self.profile != 'default':
                    if profile_name != self.profile:
                        continue
                else:
                    self.profile = profile_name
                logger.debug(f"SSO: using profile {self.profile}")
                self.sso_session = contents.get('sso_session')
                self.sso_account_id = contents.get('sso_account_id')
                self.sso_role_name = contents.get('sso_role_name')
                self.profile_region = contents.get('region')
                if not self.sso_session:
                    self.sso_start_url = contents.get('sso_start_url')
                    self.sso_region = contents.get('sso_region')
                    self.sso_registration_scopes = 'sso:account:access'
                break

        for section, contents in self.config_data.items():
            if section.startswith('sso-session'):
                session_name = section.split()[1]
                if session_name == self.sso_session:
                    self.sso_start_url = contents.get('sso_start_url')
                    self.sso_region = contents.get('sso_region')
                    self.sso_registration_scopes = contents.get('sso_registration_scopes')

    def get_auth_config(self) -> dict:
        self.test_session()
        session = botocore.session.get_session()
        return {
            'AWS_ACCESS_KEY_ID': session.get_credentials().access_key,
            'AWS_SECRET_ACCESS_KEY': session.get_credentials().secret_key,
            'AWS_SESSION_TOKEN': session.get_credentials().token,
        }

    @staticmethod
    def auth_expired(timestamp):
        if timestamp:
            _timestamp = timestamp / 1000
            expires = datetime.fromtimestamp(_timestamp).astimezone(timezone.utc)
            if datetime.now(timezone.utc) < expires:
                return False
        return True

    def default_auth(self):
        if 'AWS_ACCESS_KEY_ID' in os.environ and 'AWS_SECRET_ACCESS_KEY' in os.environ:
            self.access_key = os.environ['AWS_ACCESS_KEY_ID']
            self.secret_key = os.environ['AWS_SECRET_ACCESS_KEY']
            if 'AWS_SESSION_TOKEN' in os.environ:
                self.token = os.environ['AWS_SESSION_TOKEN']
        else:
            session = boto3.Session(profile_name=self.profile)
            credentials = session.get_credentials()
            self.access_key = credentials.access_key
            self.secret_key = credentials.secret_key

    def save_auth(self):
        dt = datetime.fromtimestamp(self.token_expiration / 1000)
        token_expiration = dt.strftime('%Y-%m-%dT%H:%M:%S.%f')
        self.credential_data[self.profile] = {
            'aws_access_key_id': self.access_key,
            'aws_secret_access_key': self.secret_key,
            'aws_session_token': self.token,
            'x_security_token_expires': token_expiration
        }
        with open(self.credential_file, 'w') as config_file:
            self.credential_data.write(config_file)

    def sso_auth(self):
        token = {}

        self.read_sso_config()

        session = botocore.session.Session()
        profiles_config = session.full_config.get('profiles', {})
        default_config = profiles_config.get(self.profile, {})

        self.access_key = default_config.get("aws_access_key_id")
        self.secret_key = default_config.get("aws_secret_access_key")
        self.token = default_config.get("aws_session_token")
        token_expiration_str = default_config.get("x_security_token_expires")
        if token_expiration_str:
            dt = datetime.strptime(token_expiration_str, '%Y-%m-%dT%H:%M:%S.%f')
            self.token_expiration = dt.timestamp() * 1000

        if not self.auth_expired(self.token_expiration):
            return

        if not self.sso_account_id or not self.sso_start_url or not self.sso_region:
            AWSDriverError("Please run \"aws configure sso\" to setup SSO")

        session = boto3.Session()
        account_id = self.sso_account_id
        start_url = self.sso_start_url
        region = self.sso_region
        sso_oidc = session.client('sso-oidc', region_name=region)
        client_creds = sso_oidc.register_client(
            clientName='couch-formation',
            clientType='public',
        )
        device_authorization = sso_oidc.start_device_authorization(
            clientId=client_creds['clientId'],
            clientSecret=client_creds['clientSecret'],
            startUrl=start_url,
        )
        url = device_authorization['verificationUriComplete']
        device_code = device_authorization['deviceCode']
        expires_in = device_authorization['expiresIn']
        interval = device_authorization['interval']

        logger.info(f"If a browser window does not open, follow this URL to continue: {url}")
        webbrowser.open_new_tab(url)

        for n in range(1, expires_in // interval + 1):
            time.sleep(interval)
            try:
                token = sso_oidc.create_token(
                    grantType='urn:ietf:params:oauth:grant-type:device_code',
                    deviceCode=device_code,
                    clientId=client_creds['clientId'],
                    clientSecret=client_creds['clientSecret'],
                )
                break
            except sso_oidc.exceptions.AuthorizationPendingException:
                pass

        access_token = token['accessToken']
        sso = session.client('sso', region_name=region)
        account_roles = sso.list_account_roles(
            accessToken=access_token,
            accountId=account_id,
        )
        roles = account_roles['roleList']
        role = next((r for r in roles if r.get('roleName') == self.sso_role_name), None)
        if not role:
            AWSDriverError(f"Role {self.sso_role_name} is not available for account {self.sso_account_id}")
        role_creds = sso.get_role_credentials(
            roleName=role['roleName'],
            accountId=account_id,
            accessToken=access_token,
        )

        session_creds = role_creds['roleCredentials']

        self.access_key = session_creds['accessKeyId']
        self.secret_key = session_creds['secretAccessKey']
        self.token = session_creds['sessionToken']
        self.token_expiration = session_creds['expiration']
        self.save_auth()

    @property
    def region(self):
        return self.aws_region

    @staticmethod
    def tag_exists(key, tags):
        for i in range(len(tags)):
            if tags[i]['Key'] == key:
                return True
        return False

    @staticmethod
    def get_tag(key, tags):
        for i in range(len(tags)):
            if tags[i]['Key'] == key:
                return tags[i]['Value']
        return None

    def get_all_regions(self) -> list:
        regions = self.ec2_client.describe_regions(AllRegions=False)
        region_list = list(r['RegionName'] for r in regions['Regions'])
        return region_list

    def zones(self) -> list:
        try:
            zone_list = self.ec2_client.describe_availability_zones()
        except Exception as err:
            raise AWSDriverError(f"error getting availability zones: {err}")

        for availability_zone in zone_list['AvailabilityZones']:
            self.zone_list.append(availability_zone['ZoneName'])

        self.zone_list = sorted(set(self.zone_list))

        if len(self.zone_list) == 0:
            raise AWSDriverError("can not get AWS availability zones")

        return self.zone_list



##
##

import re
import os
import logging
import random
import string
import uuid
from itertools import cycle
from typing import List
from couchformation.network import NetworkDriver
from couchformation.gcp.driver.network import Network, Subnet
from couchformation.gcp.driver.firewall import Firewall
from couchformation.gcp.driver.dns import DNS
from couchformation.gcp.driver.base import CloudBase
import couchformation.gcp.driver.constants as C
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, PortSettings, State
from couchformation.deployment import MetadataManager
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, synchronize, dump_class_variables

logger = logging.getLogger('couchformation.gcp.network')
logger.addHandler(logging.NullHandler())


class GCPNetworkError(FatalError):
    pass


class GCPNetwork(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.region = parameters.get('region')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.cloud = parameters.get('cloud')
        self.domain = parameters.get('domain')
        self.peer_project = parameters.get('peer_gcp_project')
        self.peer_network = parameters.get('peer_gcp_network')
        self.managed_zone = parameters.get('managed_gcp_zone')
        self.allow = parameters.get('allow') if parameters.get('allow') else "0.0.0.0/0"
        self.build_ports = PortSettingSet().create().items()

        filename = get_state_file(self.project, f"network-{self.region}")

        try:
            state_dir = get_state_dir(self.project, f"network-{self.region}")
            if not os.path.exists(state_dir):
                FileManager().make_dir(state_dir)
        except Exception as err:
            raise GCPNetworkError(f"can not create state dir: {err}")

        document = f"network:{self.cloud}"
        self.state = KeyValueStore(filename, document)

        self.gcp_network = Network(self.parameters)
        self.gcp_base = CloudBase(self.parameters)

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.vpc_name = f"{self.asset_prefix}-vpc"
        self.subnet_name = f"{self.asset_prefix}-subnet-01"
        self.firewall_default = f"{self.vpc_name}-fw-default"
        self.firewall_ssh = f"{self.vpc_name}-fw-ssh"
        self.firewall_win = f"{self.vpc_name}-fw-win"

        if self.peer_network and self.peer_project:
            peer_name_part = uuid.uuid5(uuid.NAMESPACE_URL, self.peer_project + self.peer_network)
            self.peer_name = f"cf-{peer_name_part}-peer"
            self.peer_managed_zone_name = f"cf-{peer_name_part}-zone"

    def check_state(self):
        if self.state.get('firewall_win'):
            result = Firewall(self.parameters).details(self.state['firewall_win'])
            if result is None:
                logger.warning(f"Removing stale state entry for firewall entry {self.state['firewall_win']}")
                del self.state['firewall_win']
        else:
            result = Firewall(self.parameters).details(self.firewall_win)
            if result:
                logger.warning(f"Importing orphaned entry for firewall rule {self.firewall_win}")
                self.state['firewall_win'] = self.firewall_win

        if self.state.get('firewall_ssh'):
            result = Firewall(self.parameters).details(self.state['firewall_ssh'])
            if result is None:
                logger.warning(f"Removing stale state entry for firewall entry {self.state['firewall_ssh']}")
                del self.state['firewall_ssh']
        else:
            result = Firewall(self.parameters).details(self.firewall_ssh)
            if result:
                logger.warning(f"Importing orphaned entry for firewall rule {self.firewall_ssh}")
                self.state['firewall_ssh'] = self.firewall_ssh

        for build_port_cfg in self.build_ports:
            build_name = build_port_cfg.build
            state_key_name = f"firewall_{build_name}"
            build_fw_name = f"{self.vpc_name}-fw-{build_name}"
            if self.state.get(state_key_name):
                result = Firewall(self.parameters).details(self.state[state_key_name])
                if result is None:
                    logger.warning(f"Removing stale state entry for firewall entry {self.state[state_key_name]}")
                    del self.state[state_key_name]
            else:
                result = Firewall(self.parameters).details(build_fw_name)
                if result:
                    logger.warning(f"Importing orphaned entry for firewall rule {build_fw_name}")
                    self.state[state_key_name] = build_fw_name

        for group_sg_key in self.state.key_match('firewall_.*_group_.*'):
            if self.state.get(group_sg_key):
                result = Firewall(self.parameters).details(self.state[group_sg_key])
                if result is None:
                    logger.warning(f"Removing stale state entry for firewall entry {self.state[group_sg_key]}")
                    del self.state[group_sg_key]
        for fw_entry in Firewall(self.parameters).search(f"{self.vpc_name}-fw-.*-[0-9]*"):
            m = re.search(f"{self.vpc_name}-fw-(.+?)-(.+?)", fw_entry['name'])
            state_key_name = f"firewall_{m.group(1)}_group_{m.group(2)}"
            if not self.state.get(state_key_name):
                logger.warning(f"Importing orphaned entry for firewall rule {fw_entry['name']}")
                self.state[state_key_name] = fw_entry['name']

        if self.state.get('firewall_default'):
            result = Firewall(self.parameters).details(self.state['firewall_default'])
            if result is None:
                logger.warning(f"Removing stale state entry for firewall entry {self.state['firewall_default']}")
                del self.state['firewall_default']
        else:
            result = Firewall(self.parameters).details(self.firewall_default)
            if result:
                logger.warning(f"Importing orphaned entry for firewall rule {self.firewall_default}")
                self.state['firewall_default'] = self.firewall_default

        if self.state.get('subnet'):
            result = Subnet(self.parameters).details(self.state['subnet'])
            if result is None:
                logger.warning(f"Removing stale state entry for subnet {self.state['subnet']}")
                del self.state['subnet']
                del self.state['subnet_cidr']
        else:
            result = Subnet(self.parameters).details(self.subnet_name)
            if result:
                logger.warning(f"Importing orphaned entry for subnet {self.subnet_name}")
                self.state['subnet'] = self.subnet_name
                self.state['subnet_cidr'] = result['cidr']

        if self.state.get('network'):
            result = Network(self.parameters).details(self.state['network'])
            if result is None:
                logger.warning(f"Removing stale state entry for network {self.state['network']}")
                del self.state['network']
                del self.state['network_cidr']
                del self.state['zone']
        else:
            result = Network(self.parameters).details(self.vpc_name)
            if result:
                logger.warning(f"Importing orphaned entry for network {self.vpc_name}")
                self.state['network'] = self.vpc_name
                self.state['network_link'] = result['selfLink']

        if self.state.get('public_hosted_zone'):
            result = DNS(self.parameters).details(self.state['public_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for public managed zone {self.state['public_hosted_zone']}")
                del self.state['public_hosted_zone']

        if self.state.get('private_hosted_zone'):
            result = DNS(self.parameters).details(self.state['private_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for private managed zone {self.state['private_hosted_zone']}")
                del self.state['private_hosted_zone']

    @synchronize()
    def create_vpc(self):
        self.check_state()
        cidr_util = NetworkDriver()

        for net in self.gcp_network.cidr_list:
            cidr_util.add_network(net)

        zone_list = self.gcp_network.zones()

        try:
            self.state['state'] = State.DEPLOYING.value
            logger.debug(f"GCP network create input variables:\n{dump_class_variables(vars(self))}")

            if not self.state.get('project_name'):
                self.state['project_name'] = self.gcp_base.project

            if not self.state.get('project_number'):
                self.state['project_number'] = self.gcp_base.project_number

            if not self.state.get('default_service_account'):
                self.state['default_service_account'] = self.gcp_base.default_sa

            logger.info(f"Creating network in project {self.gcp_base.project} ({self.gcp_base.project_number})")
            logger.info(f"Network region: {self.region}")

            if not self.state.get('network'):
                vpc_cidr = cidr_util.get_next_network()
                network_link = Network(self.parameters).create(self.vpc_name)
                self.state['network'] = self.vpc_name
                self.state['network_cidr'] = vpc_cidr
                self.state['network_link'] = network_link
                logger.info(f"Created network {self.vpc_name}")
                logger.info(f"Network CIDR: {vpc_cidr}")
            else:
                self.vpc_name = self.state['network']
                vpc_cidr = self.state['network_cidr']
                network_link = self.state['network_link']
                cidr_util.set_active_network(vpc_cidr)

            subnet_list = list(cidr_util.get_next_subnet())
            subnet_cycle = cycle(subnet_list[1:])

            if not self.state.get('subnet_cidr'):
                subnet_cidr = next(subnet_cycle)
                self.state['subnet_cidr'] = subnet_cidr
            else:
                subnet_cidr = self.state['subnet_cidr']

            if not self.state.get('subnet'):
                Subnet(self.parameters).create(self.subnet_name, self.vpc_name, subnet_cidr)
                self.state['subnet'] = self.subnet_name
                logger.info(f"Created subnet {self.subnet_name}")

            if not self.state.get('firewall_default'):
                Firewall(self.parameters).create_ingress(self.firewall_default, self.vpc_name, vpc_cidr, "all")
                self.state['firewall_default'] = self.firewall_default
                logger.info(f"Created firewall rule {self.firewall_default}")

            if not self.state.get('firewall_ssh'):
                Firewall(self.parameters).create_ingress(self.firewall_ssh, self.vpc_name, "0.0.0.0/0", "tcp", ["22"])
                self.state['firewall_ssh'] = self.firewall_ssh
                logger.info(f"Created firewall rule {self.firewall_ssh}")

            for n, zone in enumerate(zone_list):
                if self.state.list_exists('zone', zone):
                    continue
                self.state.list_add('zone', zone, self.subnet_name)
                logger.info(f"Added zone {zone}")

            if self.domain and not self.state.get('domain'):
                domain_prefix = ''.join(random.choice(string.ascii_lowercase) for _ in range(7))
                domain_name = f"{domain_prefix}.{self.domain}"
                self.state['domain'] = domain_name
                logger.info(f"Generated project domain {domain_name}")
            elif self.state.get('domain'):
                domain_name = self.state.get('domain')
                logger.info(f"Using existing domain {domain_name}")
            else:
                domain_name = None

            if domain_name and not self.state.get('public_hosted_zone'):
                domain_id = DNS(self.parameters).create(domain_name, private=False)
                self.state['public_hosted_zone'] = domain_id
                logger.info(f"Created public managed zone {domain_id} for domain {domain_name}")

            if domain_name and not self.state.get('private_hosted_zone'):
                domain_id = DNS(self.parameters).create(domain_name, network_link, private=True)
                self.state['private_hosted_zone'] = domain_id
                logger.info(f"Created private managed zone {domain_id} for domain {domain_name}")

            if self.state.get('public_hosted_zone') and not self.state['parent_hosted_zone']:
                parent_domain = '.'.join(domain_name.split('.')[1:])
                parent_id = DNS(self.parameters).zone_name(parent_domain)
                if parent_id:
                    ns_names = DNS(self.parameters).record_sets(self.state['public_hosted_zone'], 'NS')
                    DNS(self.parameters).add_record(parent_id, domain_name, ns_names, 'NS')
                    self.state['parent_hosted_zone'] = parent_id
                    self.state['parent_zone_ns_records'] = ','.join(ns_names)
                    logger.info(f"Added {len(ns_names)} NS record(s) to domain {parent_domain}")

            self.state['state'] = State.DEPLOYED.value
        except Exception as err:
            raise GCPNetworkError(f"Error creating network: {err}")

    @synchronize()
    def create_build_sg(self, build_name: str):
        vpc_name = self.vpc_name
        for build_port_cfg in self.build_ports:
            if build_port_cfg.build != build_name:
                continue
            state_key_name = f"firewall_{build_name}"
            build_fw_name = f"{vpc_name}-fw-{build_name}"
            if not self.state.get(state_key_name):
                Firewall(self.parameters).create_ingress(build_fw_name, vpc_name, self.allow,
                                                         ports=list(build_port_cfg.tcp_as_ranges()),
                                                         udp_ports=list(build_port_cfg.udp_as_ranges()))
                self.state[state_key_name] = build_fw_name
                logger.info(f"Created firewall rule {build_fw_name}")

    @synchronize()
    def create_win_sg(self):
        vpc_name = self.vpc_name
        if not self.state.get('firewall_win'):
            win_fw_name = f"{self.vpc_name}-fw-win"
            Firewall(self.parameters).create_ingress(win_fw_name, vpc_name, self.allow, "tcp", [
                "3389",
                "5985",
                "5986"
            ])
            self.state['firewall_win'] = win_fw_name
            logger.info(f"Created firewall rule {win_fw_name}")

    @synchronize()
    def create_node_group_sg(self, service: str, group: int, ports: List[str]):
        vpc_name = self.vpc_name
        state_key_name = f"firewall_{service}_group_{group}"
        build_fw_name = f"{vpc_name}-fw-{service}-{group}"
        if not self.state.get(state_key_name):
            port_cfg = PortSettings().create(self.name, ports)
            Firewall(self.parameters).create_ingress(build_fw_name, vpc_name, self.allow, ports=list(port_cfg.tcp_as_ranges()), udp_ports=list(port_cfg.udp_as_ranges()))
            self.state[state_key_name] = build_fw_name
            logger.info(f"Created firewall rule {build_fw_name}")

    @synchronize()
    def peer_vpc(self):
        if not self.peer_project or not self.peer_network:
            return
        self.check_state()

        Network(self.parameters).add_peering(self.peer_name, self.vpc_name, self.peer_project, self.peer_network)
        logger.info(f"Added peering to network {self.peer_network} in project {self.peer_project}")

        if self.managed_zone:
            net_link = self.state['network_link']
            service_account = self.gcp_base.default_sa
            DNS(self.parameters).create(self.managed_zone, net_link, True, self.peer_project, self.peer_network, service_account, self.peer_managed_zone_name)
            logger.info(f"Created managed zone {self.peer_managed_zone_name} ({self.managed_zone})")

    @synchronize()
    def unpeer_vpc(self):
        if not self.peer_project or not self.peer_network:
            return

        Network(self.parameters).remove_peering(self.peer_name, self.vpc_name)
        logger.info(f"Removed peering {self.peer_name}")
        DNS(self.parameters).delete(self.peer_managed_zone_name)
        logger.info(f"Removed managed zone {self.peer_managed_zone_name}")

    @synchronize()
    def destroy_vpc(self):
        if self.state.list_len('services') > 0:
            logger.info(f"Active services, leaving project network in place")
            return

        self.check_state()
        try:
            self.state['state'] = State.DESTROYING.value

            if self.state.get('firewall_win'):
                firewall_win = self.state.get('firewall_win')
                Firewall(self.parameters).delete(firewall_win)
                del self.state['firewall_win']
                logger.info(f"Removed firewall rule {firewall_win}")

            if self.state.get('firewall_ssh'):
                firewall_ssh = self.state.get('firewall_ssh')
                Firewall(self.parameters).delete(firewall_ssh)
                del self.state['firewall_ssh']
                logger.info(f"Removed firewall rule {firewall_ssh}")

            for build_port_cfg in self.build_ports:
                build_name = build_port_cfg.build
                state_key_name = f"firewall_{build_name}"
                if self.state.get(state_key_name):
                    firewall_rule = self.state.get(state_key_name)
                    Firewall(self.parameters).delete(firewall_rule)
                    del self.state[state_key_name]
                    logger.info(f"Removing firewall rule {firewall_rule}")

            for group_sg_key in self.state.key_match('firewall_.*_group_.*'):
                if self.state.get(group_sg_key):
                    firewall_rule = self.state.get(group_sg_key)
                    Firewall(self.parameters).delete(firewall_rule)
                    del self.state[group_sg_key]
                    logger.info(f"Removing firewall rule {firewall_rule}")

            if self.state.get('firewall_default'):
                firewall_default = self.state.get('firewall_default')
                Firewall(self.parameters).delete(firewall_default)
                del self.state['firewall_default']
                logger.info(f"Removed firewall rule {firewall_default}")

            if self.state.get('subnet'):
                subnet_name = self.state.get('subnet')
                Subnet(self.parameters).delete(subnet_name)
                del self.state['subnet']
                logger.info(f"Removed subnet {subnet_name}")

            if self.state.get('subnet_cidr'):
                del self.state['subnet_cidr']

            for n, zone_state in reversed(list(enumerate(self.state.list_get('zone')))):
                self.state.list_remove('zone', zone_state[0])

            if self.state.get('parent_hosted_zone') and self.state.get('domain'):
                DNS(self.parameters).delete_record(self.state['parent_hosted_zone'], self.state['domain'], 'NS')
                del self.state['parent_hosted_zone']
                del self.state['parent_zone_ns_records']
                logger.info(f"Removing NS records for domain {self.state['domain']}")

            if self.state.get('public_hosted_zone'):
                domain_id = self.state.get('public_hosted_zone')
                DNS(self.parameters).delete(domain_id)
                del self.state['public_hosted_zone']
                logger.info(f"Removing public hosted zone {domain_id}")

            if self.state.get('private_hosted_zone'):
                domain_id = self.state.get('private_hosted_zone')
                DNS(self.parameters).delete(domain_id)
                del self.state['private_hosted_zone']
                logger.info(f"Removing private hosted zone {domain_id}")

            if self.state.get('domain'):
                domain_name = self.state.get('domain')
                del self.state['domain']
                logger.info(f"Removing project domain {domain_name}")

            if self.state.get('network'):
                vpc_name = self.state.get('network')
                Network(self.parameters).delete(vpc_name)
                del self.state['network']
                del self.state['network_cidr']
                del self.state['network_link']
                logger.info(f"Removed network {vpc_name}")

            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise GCPNetworkError(f"Error removing network: {err}")

    def create(self):
        logger.info(f"Creating cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.create_vpc()

    def destroy(self):
        logger.info(f"Removing cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.destroy_vpc()

    def get(self, key):
        return self.state.get(key)

    @property
    def network(self):
        return self.state.get('network')

    @property
    def subnet(self):
        return self.state.get('subnet')

    @property
    def zones(self):
        return self.state.list_get('zone')

    @property
    def domain_name(self):
        return self.state.get('domain')

    @property
    def public_zone(self):
        return self.state.get('public_hosted_zone')

    @property
    def private_zone(self):
        return self.state.get('private_hosted_zone')

    def add_service(self, name):
        self.state.list_add('services', name)

    def remove_service(self, name):
        self.state.list_remove('services', name)



##
##

import logging
import googleapiclient.errors
import googleapiclient.discovery
from couchformation.gcp.driver.base import CloudBase, GCPDriverError

logger = logging.getLogger('couchformation.gcp.driver.dns')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class DNS(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def fqdn(domain: str):
        if domain[-1] not in ['.']:
            domain = domain + '.'
        return domain

    def create(self,
               domain: str,
               network_link: str = None,
               private: bool = False,
               peer_project: str = None,
               peer_network: str = None,
               service_account: str = None,
               zone_name: str = None):
        if service_account:
            dns_client = googleapiclient.discovery.build('dns', 'v1', credentials=self.sa_auth(service_account))
        else:
            dns_client = googleapiclient.discovery.build('dns', 'v1', credentials=self.credentials)

        if zone_name:
            name = zone_name
        else:
            name_part = domain.replace('.', '-')
            name = f"{name_part}-public" if not private else f"{name_part}-private"

        visibility = 'private' if private else 'public'

        dns_body = {
            'kind': 'dns#managedZone',
            'name': name,
            'dnsName': self.fqdn(domain),
            'description': 'Couch Formation Managed Zone',
            'visibility': visibility
        }

        if private and network_link:
            dns_body['privateVisibilityConfig'] = {
                "kind": "dns#managedZonePrivateVisibilityConfig",
                "networks": [
                    {
                        "kind": "dns#managedZonePrivateVisibilityConfigNetwork",
                        "networkUrl": network_link
                    }
                ]
            }

        if peer_project and peer_network:
            dns_body['peeringConfig'] = {
                "targetNetwork": {
                    "networkUrl": f"https://www.googleapis.com/compute/v1/projects/{peer_project}/global/networks/{peer_network}",
                    "kind": "dns#managedZonePeeringConfigTargetNetwork"
                },
                "kind": "dns#managedZonePeeringConfig"
            }

        try:
            request = dns_client.managedZones().create(project=self.gcp_project, body=dns_body)
            result = request.execute()
            return result.get('name')
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create managed zone: {err}")
            return name
        except Exception as err:
            raise GCPDriverError(f"error creating managed zone: {err}")

    def details(self, name: str):
        try:
            request = self.dns_client.managedZones().get(project=self.gcp_project, managedZone=name)
            result = request.execute()
            return result
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find managed zone: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting managed zone: {err}")

    def list_zones(self):
        zone_list = []
        try:
            request = self.dns_client.managedZones().list(project=self.gcp_project)
            while request is not None:
                response = request.execute()
                for managed_zone in response['managedZones']:
                    zone_list.append(managed_zone)
                request = self.dns_client.managedZones().list_next(previous_request=request, previous_response=response)
            return zone_list
        except Exception as err:
            raise GCPDriverError(f"error listing managed zones: {err}")

    def zone_name(self, domain: str):
        zones = self.list_zones()
        return next((zone['name'] for zone in zones if zone['dnsName'].startswith(domain)), None)

    def record_sets(self, name: str, r_type: str):
        record_list = []
        try:
            request = self.dns_client.resourceRecordSets().list(project=self.gcp_project, managedZone=name)
            while request is not None:
                response = request.execute()
                for resource_record_set in response['rrsets']:
                    if resource_record_set['type'] != r_type:
                        continue
                    record_list.extend(resource_record_set['rrdatas'])
                request = self.dns_client.managedZones().list_next(previous_request=request, previous_response=response)
            return record_list
        except Exception as err:
            raise GCPDriverError(f"error listing managed zones: {err}")

    def delete(self, name: str):
        try:
            request = self.dns_client.managedZones().delete(project=self.gcp_project, managedZone=name)
            request.execute()
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not delete managed zone: {err}")
        except Exception as err:
            raise GCPDriverError(f"error deleting managed zone: {err}")

    def add_record(self, managed_zone: str, name: str, values: list, record_type: str = 'A', ttl: int = 300):
        dns_record_body = {
            'kind': 'dns#resourceRecordSet',
            'name': self.fqdn(name),
            'rrdatas': values,
            'ttl': ttl,
            'type': record_type
        }

        try:
            request = self.dns_client.resourceRecordSets().create(project=self.gcp_project, managedZone=managed_zone, body=dns_record_body)
            request.execute()
        except Exception as err:
            raise GCPDriverError(f"error creating DNS records: {err}")

    def delete_record(self, managed_zone: str, name: str, record_type: str = 'A'):
        try:
            request = self.dns_client.resourceRecordSets().delete(project=self.gcp_project, managedZone=managed_zone, name=self.fqdn(name), type=record_type)
            request.execute()
        except Exception as err:
            raise GCPDriverError(f"error deleting DNS records: {err}")



##
##

from typing import List, Optional, Dict
from enum import Enum
import logging
import ast
import attr
import yaml
import argparse
import couchformation.constants as C

logger = logging.getLogger('couchformation.executor.targets')
logger.addHandler(logging.NullHandler())


class DeployMode(Enum):
    node = 'node'
    saas = 'saas'


@attr.s
class Profile:
    driver: str = attr.ib()
    module: str = attr.ib()
    deploy: str = attr.ib()
    destroy: str = attr.ib()
    peer: str = attr.ib()
    unpeer: str = attr.ib()
    info: str = attr.ib()
    compose: str = attr.ib()


@attr.s
class Parameters:
    options: List[str] = attr.ib()
    required: Optional[List[str]] = attr.ib(default=[])
    boolean: Optional[List[str]] = attr.ib(default=[])


@attr.s
class CloudDriverBase:
    driver: str = attr.ib()
    module: str = attr.ib()
    test: str = attr.ib()


@attr.s
class CloudProfile:
    name: str = attr.ib()
    base: CloudDriverBase = attr.ib()
    network: Profile = attr.ib()
    node: Profile = attr.ib()
    parameters: Parameters = attr.ib()
    options: Optional[argparse.Namespace] = attr.ib(default=argparse.Namespace())
    undefined: Optional[List[str]] = attr.ib(default=[])
    parser: Optional[argparse.ArgumentParser] = attr.ib(default=argparse.ArgumentParser(add_help=False))

    def check_required_options(self):
        missing = []
        for opt in self.parameters.required:
            if opt not in self.options or getattr(self.options, opt) is None:
                missing.append(opt)
        return missing

    def check_undefined_options(self):
        return self.undefined

    def get_options(self):
        return self.options

    def merge_options(self, parameters: dict):
        p_copy = parameters.copy()
        opt_data = vars(self.options)
        for key, value in opt_data.items():
            if key not in parameters or parameters[key] is None:
                p_copy[key] = value
        return p_copy


@attr.s
class ProfileSet:
    profiles: List[CloudProfile] = attr.ib(default=[])

    def add(self, p: CloudProfile):
        self.profiles.append(p)

    def get(self, cloud):
        return next((p for p in self.profiles if p.name == cloud), None)


@attr.s
class BuildConfig:
    provisioner: str = attr.ib()
    root: bool = attr.ib()
    commands: List[str] = attr.ib()
    exclude: Optional[List[str]] = attr.ib(default=[])
    files: Optional[dict] = attr.ib(default={})


@attr.s
class BuildConfigSequence:
    name: str = attr.ib()
    sequence: List[BuildConfig] = attr.ib(default=[])

    def add(self, b: BuildConfig):
        self.sequence.append(b)

    def get(self, provisioner) -> List[BuildConfig]:
        return list(s for s in self.sequence if s.provisioner == provisioner)


@attr.s
class BuildSet:
    profiles: List[BuildConfigSequence] = attr.ib(default=[])

    def add(self, p: BuildConfigSequence):
        self.profiles.append(p)

    def get(self, name):
        return next((p for p in self.profiles if p.name == name), None)


@attr.s
class Provisioner:
    name: str = attr.ib()
    driver: str = attr.ib()
    module: str = attr.ib()
    method: str = attr.ib()
    upload: str = attr.ib()
    files: str = attr.ib()
    when: str = attr.ib()
    options: List[str] = attr.ib()
    parameters: Dict = attr.ib()

    def parameter_gen(self, *args):
        parameters = self.initialize_parameters(self.options, args)
        return parameters

    @staticmethod
    def initialize_parameters(expected, args):
        p = {}
        for d in args:
            for attribute in expected:
                if p.get(attribute):
                    continue
                if d.get(attribute):
                    p[attribute] = d.get(attribute)
        return p


@attr.s
class ProvisionerSet:
    provisioners: List[Provisioner] = attr.ib(default=[])

    def add(self, p: Provisioner):
        self.provisioners.append(p)

    def get(self, name):
        return next((p for p in self.provisioners if p.name == name), None)


@attr.s
class Strategy:
    name: str = attr.ib()
    deployer: str = attr.ib()


@attr.s
class StrategySet:
    strategies: List[Strategy] = attr.ib(default=[])

    def add(self, s: Strategy):
        self.strategies.append(s)

    def get(self, name):
        return next((p for p in self.strategies if p.name == name), None)


##
@attr.s
class CloudType:
    cloud: str = attr.ib()
    provisioner: str = attr.ib()


@attr.s
class CloudTypeSet:
    clouds: List[CloudType] = attr.ib(default=[])

    def add(self, c: CloudType):
        self.clouds.append(c)

    def get(self, name):
        return next((c for c in self.clouds if c.cloud == name), None)


class CloudProfileBase(object):

    def __init__(self, cloud: str):
        self.cloud = cloud
        cfg_file = C.TARGET_PROFILES
        with open(cfg_file, "r") as f:
            try:
                for cloud_name, settings in yaml.safe_load(f).items():
                    if cloud == cloud_name:
                        base = CloudDriverBase(*self.construct_driver(settings, 'base'))
                        network = Profile(*self.construct_profile(settings, 'network'))
                        node = Profile(*self.construct_profile(settings, 'node'))
                        self._profile = CloudProfile(cloud, base, network, node, Parameters(settings.get('parameters'), settings.get('required', []), settings.get('boolean', [])))
                        break
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open target config file {cfg_file}: {err}")

    @property
    def profile(self) -> CloudProfile:
        if not self._profile:
            raise ValueError(f"Cloud {self.cloud} is not supported")
        return self._profile

    @staticmethod
    def construct_profile(settings, key):
        config = settings.get(key)
        driver = list(config.keys())[0]
        elements = config.get(driver)
        module = elements.get('module')
        deploy = elements.get('deploy')
        destroy = elements.get('destroy')
        peer = elements.get('peer')
        unpeer = elements.get('unpeer')
        info = elements.get('info')
        compose = elements.get('compose')
        return driver, module, deploy, destroy, peer, unpeer, info, compose

    @staticmethod
    def construct_driver(settings, key):
        config = settings.get(key)
        driver = list(config.keys())[0]
        elements = config.get(driver)
        module = elements.get('module')
        test = elements.get('test')
        return driver, module, test


class TargetProfile(object):

    def __init__(self, options):
        self.cfg_file = C.TARGET_PROFILES
        self.options = options
        self.config = ProfileSet()
        self.load_config()

    def get(self, cloud) -> CloudProfile:
        profile = self.config.get(cloud)
        if not profile:
            raise ValueError(f"Cloud {cloud} is not supported")
        profile.options, profile.undefined, profile.parser = self.initialize_args(self.options, profile.parameters.options)
        return profile

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for cloud, settings in yaml.safe_load(f).items():
                    base = CloudDriverBase(*self.construct_driver(settings, 'base'))
                    network = Profile(*self.construct_profile(settings, 'network'))
                    node = Profile(*self.construct_profile(settings, 'node'))
                    profile = CloudProfile(cloud, base, network, node, Parameters(settings.get('parameters'), settings.get('required', []), settings.get('boolean', [])))
                    self.config.add(profile)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open target config file {self.cfg_file}: {err}")

    @staticmethod
    def construct_profile(settings, key):
        config = settings.get(key)
        driver = list(config.keys())[0]
        elements = config.get(driver)
        module = elements.get('module')
        deploy = elements.get('deploy')
        destroy = elements.get('destroy')
        peer = elements.get('peer')
        unpeer = elements.get('unpeer')
        info = elements.get('info')
        compose = elements.get('compose')
        return driver, module, deploy, destroy, peer, unpeer, info, compose

    @staticmethod
    def construct_driver(settings, key):
        config = settings.get(key)
        driver = list(config.keys())[0]
        elements = config.get(driver)
        module = elements.get('module')
        test = elements.get('test')
        return driver, module, test

    @staticmethod
    def initialize_args(args, parameters):
        opt_struct = C.get_option_struct()
        parser = argparse.ArgumentParser(add_help=False)
        for attribute in parameters:
            opt_settings = opt_struct.get(attribute)
            if eval(opt_settings.type) is bool:
                parser.add_argument(f"--{attribute}", action='store_true', help=opt_settings.help)
            else:
                parser.add_argument(f"--{attribute}", action='store', type=eval(opt_settings.type), help=opt_settings.help)
        options, undefined = parser.parse_known_args(args)
        return options, undefined, parser


class BuildProfile(object):

    def __init__(self):
        self.cfg_file = C.NODE_PROFILES
        self.config = BuildSet()
        self.load_config()

    def get(self, name) -> BuildConfigSequence:
        sequence = self.config.get(name)
        if not sequence:
            return BuildConfigSequence(name, [])
        return sequence

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for name, settings in yaml.safe_load(f).items():
                    sequence = BuildConfigSequence(name, [])
                    for element in settings:
                        profile = BuildConfig(element.get('provisioner'), element.get('root'),
                                              element.get('commands'), element.get('exclude', []), element.get('files', {}))
                        sequence.add(profile)
                    self.config.add(sequence)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open node config file {self.cfg_file}: {err}")


class ProvisionerProfile(object):

    def __init__(self):
        self.cfg_file = C.PROVISIONER_PROFILES
        self.config = ProvisionerSet()
        self.load_config()

    def get(self, name) -> Provisioner:
        profile = self.config.get(name)
        if not profile:
            raise ValueError(f"Provisioner type {name} is not supported")
        return profile

    @staticmethod
    def run(parameters, expression):
        try:
            expression_node = ast.parse(expression, mode='eval')
        except SyntaxError:
            raise RuntimeError(f"provisioner profile parser: syntax error in expression {expression}")

        safe_parameters = {}
        for key, val in parameters.items():
            if isinstance(key, str) and (isinstance(val, (int, float, str, bool, list, dict))):
                safe_parameters[key] = val

        try:
            result = eval(compile(expression_node, '<string>', mode='eval'), {"__builtins__": None}, safe_parameters)
            logger.debug(f"provisioner profile: evaluated expression {expression} as {result}")
            return result
        except TypeError:
            return False
        except Exception as e:
            raise RuntimeError(f"provisioner profile parser: can not evaluate expression {expression}: {e}")

    def search(self, parameters):
        for provisioner in self.config.provisioners:
            expression = provisioner.when
            if self.run(parameters, expression):
                return provisioner.name
        return None

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for name, settings in yaml.safe_load(f).items():
                    provisioner = Provisioner(*self.construct_profile(name, settings))
                    self.config.add(provisioner)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open provisioner config file {self.cfg_file}: {err}")

    @staticmethod
    def construct_profile(name, settings):
        driver = settings.get('driver')
        module = settings.get('module')
        method = settings.get('method')
        upload = settings.get('upload')
        files = settings.get('files')
        when = settings.get('when')
        options = settings.get('parameters')
        parameters = {}
        return name, driver, module, method, upload, files, when, options, parameters


class DeployStrategy(object):

    def __init__(self):
        self.cfg_file = C.STRATEGY_PROFILES
        self.config = StrategySet()
        self.load_config()

    def get(self, name) -> Strategy:
        strategy = self.config.get(name)
        if not strategy:
            raise ValueError(f"Strategy type {name} is not supported")
        return strategy

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for name, settings in yaml.safe_load(f).items():
                    strategy = Strategy(*self.construct_strategy(name, settings))
                    self.config.add(strategy)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open strategy config file {self.cfg_file}: {err}")

    @staticmethod
    def construct_strategy(name, settings):
        deployer = settings.get('deployer')
        return name, deployer


class CloudConfig(object):

    def __init__(self):
        self.cfg_file = C.CLOUD_PROFILES
        self.config = CloudTypeSet()
        self.load_config()

    def get(self, name) -> CloudType:
        cloud = self.config.get(name)
        if not cloud:
            raise ValueError(f"Cloud type {name} is not supported")
        return cloud

    def load_config(self):
        with open(self.cfg_file, "r") as f:
            try:
                for name, settings in yaml.safe_load(f).items():
                    cloud = CloudType(*self.construct_strategy(name, settings))
                    self.config.add(cloud)
            except yaml.YAMLError as err:
                raise RuntimeError(f"Can not open strategy config file {self.cfg_file}: {err}")

    @staticmethod
    def construct_strategy(name, settings):
        provisioner = settings.get('provisioner')
        return name, provisioner


class OptionsUtility(object):

    def __init__(self):
        pass

    @staticmethod
    def get_args_as_dict(arguments: List[str], cloud: str) -> dict:
        profile = TargetProfile(arguments).get(cloud)
        return vars(profile.options)



##
##

import logging
import ipaddress
from docker.types import IPAMPool, IPAMConfig
from docker.models.networks import Network as NetworkClass
from typing import Union, List
from couchformation.docker.driver.base import CloudBase, DockerDriverError

logger = logging.getLogger('couchformation.docker.driver.network')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class Network(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, name: str = None) -> Union[List[dict], None]:
        network_list = []

        try:
            networks = self.client.networks.list()
        except Exception as err:
            raise DockerDriverError(f"error getting network list: {err}")

        for network_entry in list(n.attrs for n in networks):
            if name and network_entry.get('Name') != name:
                continue
            if network_entry.get('Driver') != 'bridge':
                continue
            net_block = {'cidr': network_entry.get('IPAM').get('Config')[0].get('Subnet'),
                         'gateway': network_entry.get('IPAM').get('Config')[0].get('Gateway'),
                         'id': network_entry.get('Id')}
            network_list.append(net_block)

        if len(network_list) == 0:
            return None
        else:
            return network_list

    @property
    def cidr_list(self):
        try:
            for item in self.list():
                yield item['cidr']
        except TypeError:
            return iter(())

    def get_network(self, name: str):
        return next((n for n in self.client.networks.list() if n.name == name), None)

    def create(self, name: str, cidr: str) -> NetworkClass:
        if self.get_network(name):
            return self.get_network(name)
        try:
            net = ipaddress.IPv4Network(cidr)
            ipam_pool = IPAMPool(
                subnet=cidr,
                gateway=str(net[1])
            )
            ipam_config = IPAMConfig(
                pool_configs=[ipam_pool]
            )
            result = self.client.networks.create(name, driver="bridge", ipam=ipam_config)
            return result
        except Exception as err:
            raise DockerDriverError(f"error creating network: {err}")

    def delete(self, name: str) -> None:
        try:
            network_id = self.get_network(name)
            if network_id:
                network_id.remove()
        except Exception as err:
            raise DockerDriverError(f"error deleting network: {err}")



##
##

import attr
import multiprocessing
import os
import yaml
from typing import List
from couchformation import get_data_dir
from pathlib import Path

GLOBAL_LOCK = multiprocessing.Lock()

DEBUG_VERBOSE_LEVEL = 1

ROOT_DIRECTORY = os.path.join(Path.home(), '.config', 'couch-formation')
STATE_DIRECTORY = os.path.join(ROOT_DIRECTORY, 'state')
LOG_DIRECTORY = os.path.join(ROOT_DIRECTORY, 'log')
CONFIG_FILE = os.path.join(ROOT_DIRECTORY, 'config.db')
DATA_DIRECTORY = get_data_dir()
NODE_PROFILES = os.path.join(DATA_DIRECTORY, "node_profiles.yaml")
TARGET_PROFILES = os.path.join(DATA_DIRECTORY, "target_profiles.yaml")
PROVISIONER_PROFILES = os.path.join(DATA_DIRECTORY, "provisioner_profiles.yaml")
STRATEGY_PROFILES = os.path.join(DATA_DIRECTORY, "strategy_profiles.yaml")
CONTAINER_PROFILES = os.path.join(DATA_DIRECTORY, "container_profiles.yaml")
CLOUD_PROFILES = os.path.join(DATA_DIRECTORY, "cloud_profiles.yaml")
BUILD_PROFILES = os.path.join(DATA_DIRECTORY, "build_profiles.yaml")
OPTION_HELP_PROFILE = os.path.join(DATA_DIRECTORY, "option_help.yaml")
NODE_PORT_PROFILE = os.path.join(DATA_DIRECTORY, "node_ports.yaml")
PLAYBOOK_DIR = os.path.join(DATA_DIRECTORY, "playbooks")

METADATA = "metadata.db"
NETWORK = "network.db"
STATE = "state.db"
CREDENTIALS = "credentials.db"

GREY_COLOR = "\x1b[38;20m"
YELLOW_COLOR = "\x1b[33;20m"
RED_COLOR = "\x1b[31;20m"
BOLD_RED_COLOR = "\x1b[31;1m"
GREEN_COLOR = "\x1b[32;20m"
BLUE_COLOR = "\x1b[34;20m"
MAGENTA_COLOR = "\x1b[35;20m"
CYAN_COLOR = "\x1b[36;20m"
SCREEN_RESET = "\x1b[0m"
FORMAT_LEVEL = "%(levelname)s"
FORMAT_NAME = "%(name)s"
FORMAT_MESSAGE = "%(message)s"
FORMAT_LINE = "(%(filename)s:%(lineno)d)"
FORMAT_EXTRA = " [%(name)s](%(filename)s:%(lineno)d)"
FORMAT_TIMESTAMP = "%(asctime)s"
FORMAT_THREAD = "%(threadName)s"

OS_VERSION_LIST = {
    'amzn': ['2', '2023'],
    'rhel': ['8', '9'],
    'centos': ['8', '9'],
    'ol': ['8', '9'],
    'rocky': ['8', '9'],
    'fedora': ['34'],
    'sles': ['12', '15'],
    'opensuse-leap': ['15'],
    'ubuntu': ['14.04', '16.04', '18.04', '20.04', '22.04'],
    'ubuntu-pro': ['16.04', '18.04', '20.04', '22.04'],
    'debian': ['10', '11'],
    'windows': ['2019', '2022'],
    'macos': ['13', '14'],
    'arch': []
}

MACHINE_TYPES = [
        {
            "name": "2x4",
            "cpu": 2,
            "memory": 4096
        },
        {
            "name": "2x8",
            "cpu": 2,
            "memory": 8192
        },
        {
            "name": "4x8",
            "cpu": 4,
            "memory": 8192
        },
        {
            "name": "4x16",
            "cpu": 4,
            "memory": 16384
        },
        {
            "name": "8x16",
            "cpu": 8,
            "memory": 16384
        },
        {
            "name": "8x24",
            "cpu": 8,
            "memory": 24576
        },
        {
            "name": "8x32",
            "cpu": 8,
            "memory": 32768
        },
        {
            "name": "16x32",
            "cpu": 16,
            "memory": 32768
        },
        {
            "name": "16x64",
            "cpu": 16,
            "memory": 65536
        },
        {
            "name": "32x64",
            "cpu": 32,
            "memory": 65536
        },
        {
            "name": "32x128",
            "cpu": 32,
            "memory": 131072
        },
        {
            "name": "48x192",
            "cpu": 48,
            "memory": 196608
        },
        {
            "name": "64x128",
            "cpu": 64,
            "memory": 131072
        },
        {
            "name": "64x256",
            "cpu": 64,
            "memory": 262144
        },
        {
            "name": "80x320",
            "cpu": 80,
            "memory": 327680
        },
        {
            "name": "2x16",
            "cpu": 2,
            "memory": 16384
        },
        {
            "name": "4x32",
            "cpu": 4,
            "memory": 32768
        },
        {
            "name": "8x64",
            "cpu": 8,
            "memory": 65536
        },
        {
            "name": "16x128",
            "cpu": 16,
            "memory": 131072
        },
        {
            "name": "32x256",
            "cpu": 32,
            "memory": 262144
        },
        {
            "name": "48x384",
            "cpu": 48,
            "memory": 393216
        },
        {
            "name": "64x512",
            "cpu": 64,
            "memory": 524288
        },
        {
            "name": "80x640",
            "cpu": 80,
            "memory": 655360
        },
        {
            "name": "2x2",
            "cpu": 2,
            "memory": 2048
        },
        {
            "name": "4x4",
            "cpu": 4,
            "memory": 4096
        },
        {
            "name": "8x8",
            "cpu": 8,
            "memory": 8192
        },
        {
            "name": "16x16",
            "cpu": 16,
            "memory": 16384
        },
        {
            "name": "32x32",
            "cpu": 32,
            "memory": 32768
        },
        {
            "name": "48x48",
            "cpu": 48,
            "memory": 49152
        },
        {
            "name": "64x64",
            "cpu": 64,
            "memory": 65536
        },
        {
            "name": "80x80",
            "cpu": 80,
            "memory": 81920
        }
    ]

provisioners = {
    "cbs": {
        "pre_provision": [
            'curl -sfL https://raw.githubusercontent.com/mminichino/host-prep-lib/main/bin/setup.sh | sudo -E bash -s - -s -g https://github.com/mminichino/host-prep-lib',
        ],
        "provision": [
            'sudo bundlemgr -b CBS',
            'sudo swmgr cluster create -n testdb -s {{ SERVICES }} -g {{ NODE_ZONE }} -D /cbdata -l {{ PRIVATE_IP_LIST }}',
        ],
        "post_provision": [
            'sudo swmgr cluster rebalance -l {{ PRIVATE_IP_LIST }}',
        ]
    },
    "sgw": {
        "pre_provision": [
            'curl -sfL https://raw.githubusercontent.com/mminichino/host-prep-lib/main/bin/setup.sh | sudo -E bash -s - -s -g https://github.com/mminichino/host-prep-lib',
        ],
        "provision": [
            'sudo bundlemgr -b InstallSGW',
            'sudo swmgr gateway configure -l {{ CONNECT_LIST }}',
            'sudo bundlemgr -b EnableSGW'
        ],
        "post_provision": []
    }
}


@attr.s
class CloudOption:
    name: str = attr.ib()
    type: str = attr.ib()
    help: str = attr.ib()


@attr.s
class CloudOptionList:
    options: List[CloudOption] = attr.ib()

    def get(self, option: str) -> CloudOption:
        return next((o for o in self.options if o.name == option), CloudOption(option, "str", "Check documentation for details"))


def get_option_struct() -> CloudOptionList:
    options = []
    with open(OPTION_HELP_PROFILE, "r") as f:
        try:
            for option, settings in yaml.safe_load(f).items():
                option_cfg = CloudOption(option, settings.get('type_class'), settings.get('help'))
                options.append(option_cfg)
            return CloudOptionList(options)
        except yaml.YAMLError as err:
            raise RuntimeError(f"Can not open target config file {OPTION_HELP_PROFILE}: {err}")



##
##

import logging
import warnings
import json
from overrides import override
from couchformation.cli.cli import CLI
from couchformation.kvdb import KeyValueStore

warnings.filterwarnings("ignore")
logger = logging.getLogger()


class DBDumpCLI(CLI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @override()
    def local_args(self):
        self.parser.add_argument('-j', '--json', action='store_true', help="Output JSON")
        self.parser.add_argument('-R', '--regexp', action='store', help="Regexp filter", default=".*")
        self.parser.add_argument('-K', '--key', action='store', help="Key regexp filter")
        self.parser.add_argument('-V', '--value', action='store', help="Value regexp filter")

    def run(self):
        for file in self.remainder:
            db = KeyValueStore(file)
            for doc in db.doc_id_match(self.options.regexp):
                document = KeyValueStore(db.file_name, doc)
                if len(document) == 0:
                    continue
                if self.options.json:
                    print(json.dumps(document.as_dict, indent=2))
                elif self.options.key:
                    for value in document.key_match(self.options.keyexp):
                        print(value)
                elif self.options.value:
                    for value in document.value_match(self.options.value):
                        print(value)
                else:
                    print(f"Document: {doc}")
                    for key, value in document.items():
                        if not key:
                            key = ""
                        print(f"{key:<12} = {value}")


def main(args=None):
    cli = DBDumpCLI(args)
    cli.run()



##
##

from couchformation import constants as C


class Provisioners(object):

    def __init__(self, node_profile: str, config_profile: str):
        self.node_profiles = C.NODE_PROFILES
        self.playbook_dir = C.PLAYBOOK_DIR
        self.node_profile = node_profile
        self.config_profile = config_profile



##
##

import base64
import logging
import re
import time
import botocore.exceptions
from datetime import datetime, timezone
from typing import Union, List

from couchformation.aws.driver.base import CloudBase, AWSDriverError
from couchformation.aws.driver.constants import AWSEbsDisk, AWSTagStruct, EbsVolume, AWSTag, PlacementType
from couchformation.ssh import SSHUtil

logger = logging.getLogger('couchformation.aws.driver.instance')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)

WIN_USER_DATA = """<powershell>
winrm quickconfig -q -force
winrm set winrm/config/service/auth '@{Basic="true"}'
$hostname = $env:computername
$certificateThumbprint = (New-SelfSignedCertificate -DnsName "${hostname}" -CertStoreLocation Cert:\LocalMachine\My).Thumbprint
winrm create winrm/config/Listener?Address=*+Transport=HTTPS "@{Hostname=`"${hostname}`"; CertificateThumbprint=`"${certificateThumbprint}`"}"
netsh advfirewall firewall add rule name="Windows Remote Management (HTTPS-In)" dir=in action=allow protocol=TCP localport=5986
</powershell>
"""


class Instance(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self,
            name: str,
            ami: str,
            ssh_key: str,
            sg_list: Union[str, List[str]],
            subnet: str,
            zone: str,
            root_size=256,
            swap_size=16,
            swap_iops=3000,
            data_size=256,
            data_iops=3000,
            instance_type="t2.micro",
            placement: PlacementType = PlacementType.ZONE,
            host_id: str = None,
            enable_winrm: bool = False,
            tags: dict = None):
        volume_type = "gp3"
        kwargs = {}
        try:
            ami_details = self.image_details(ami)
        except Exception as err:
            raise AWSDriverError(f"error getting AMI {ami} details: {err}")
        if 'BlockDeviceMappings' not in ami_details:
            raise AWSDriverError(f"can not get details for AMI {ami}")

        root_dev = ami_details['BlockDeviceMappings'][0]['DeviceName']
        match = re.search(r"/dev/(.+?)a[0-9]*", root_dev)
        root_disk_prefix = f"/dev/{match.group(1)}"
        disk_list = [
            AWSEbsDisk.build(root_dev, EbsVolume(volume_type, root_size, 3000)).as_dict,
            AWSEbsDisk.build(f"{root_disk_prefix}b", EbsVolume(volume_type, swap_size, swap_iops)).as_dict,
            AWSEbsDisk.build(f"{root_disk_prefix}c", EbsVolume(volume_type, data_size, data_iops)).as_dict,
        ]
        i_tag = AWSTagStruct.build("instance")
        i_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                i_tag.add(AWSTag(k, str(v)))

        if placement == PlacementType.ZONE:
            placement = {"AvailabilityZone": zone}
        else:
            placement = {"Tenancy": "host"}
            if host_id:
                placement.update({"HostId": host_id})

        if enable_winrm:
            kwargs['UserData'] = WIN_USER_DATA

        if type(sg_list) is str:
            security_groups = [sg_list]
        else:
            security_groups = sg_list

        try:
            result = self.ec2_client.run_instances(BlockDeviceMappings=disk_list,
                                                   ImageId=ami,
                                                   InstanceType=instance_type,
                                                   KeyName=ssh_key,
                                                   MaxCount=1,
                                                   MinCount=1,
                                                   SecurityGroupIds=security_groups,
                                                   SubnetId=subnet,
                                                   Placement=placement,
                                                   TagSpecifications=[i_tag.as_dict],
                                                   **kwargs)
        except botocore.exceptions.ClientError as err:
            response = err.response
            message = response.get('Error', {}).get('Message')
            if message:
                parts = message.split('failure message: ')
                if len(parts) == 2:
                    decoded = self.sts_client.decode_authorization_message(EncodedMessage=parts[1])
                    decoded_message = decoded.get('DecodedMessage')
                    if decoded_message:
                        import json
                        logger.debug(f"Decoded Error:\n{json.dumps(json.loads(decoded_message), indent=2)}")
            raise AWSDriverError(f"AWS client error: {err}")
        except Exception as err:
            raise AWSDriverError(f"error running instance: {err}")

        instance_id = result['Instances'][0]['InstanceId']
        waiter = self.ec2_client.get_waiter('instance_running')
        waiter.wait(InstanceIds=[instance_id])

        return instance_id

    def list(self):
        instances = []
        extra_args = {}

        try:
            while True:
                result = self.ec2_client.describe_instances(**extra_args)
                for reservation in result.get('Reservations', []):
                    instances.extend(reservation.get('Instances', []))
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting instance list: {err}")

        return instances

    def allocate_host(self, name: str, zone: str, instance_type: str):
        host_tag = [AWSTagStruct.build("dedicated-host").add(AWSTag("Name", name)).as_dict]

        try:
            result = self.ec2_client.allocate_hosts(AvailabilityZone=zone,
                                                    InstanceType=instance_type,
                                                    AutoPlacement='on',
                                                    Quantity=1,
                                                    TagSpecifications=host_tag)
            return result['HostIds'][0]
        except Exception as err:
            raise AWSDriverError(f"error getting instance details: {err}")

    def list_hosts(self, instance_type: str = None):
        hosts = []
        host_list = []
        extra_args = {}
        machine_filter = []

        if instance_type:
            machine_filter = [
                {
                    'Name': 'instance-type',
                    'Values': [
                        instance_type,
                    ]
                }
            ]

        try:
            while True:
                result = self.ec2_client.describe_hosts(**extra_args, Filters=machine_filter)
                hosts.extend(result['Hosts'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting instance details: {err}")

        for host in hosts:
            difference = datetime.now(timezone.utc) - host['AllocationTime']
            age = int(difference.total_seconds() / 3600)
            host_block = {'id': host['HostId'],
                          'state': host['State'],
                          'created': host['AllocationTime'],
                          'age': age,
                          'zone': host['AvailabilityZone'],
                          'capacity': host.get('AvailableCapacity', {}).get('AvailableVCpus', 0),
                          'instances': [i['InstanceId'] for i in host['Instances']],
                          'machine': host['HostProperties']['InstanceType']}
            host_list.append(host_block)

        return host_list

    def get_host_by_instance(self, instance_id: str):
        host_list = self.list_hosts()
        return next((h for h in host_list if instance_id in h['instances']), None)

    def get_host_by_id(self, host_id: str):
        host_list = self.list_hosts()
        return next((h for h in host_list if h['id'] == host_id), None)

    def release_host(self, host_id: str):
        try:
            result = self.ec2_client.release_hosts(HostIds=[host_id])
            if 'Unsuccessful' in result and len(result['Unsuccessful']) > 0:
                raise AWSDriverError(f"Can not release host {host_id}: {result['Unsuccessful'][0]['Error']['Message']}")
        except Exception as err:
            raise AWSDriverError(f"error getting instance details: {err}")

    def details(self, instance_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_instances(InstanceIds=[instance_id])
            return result['Reservations'][0]['Instances'][0]
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting instance details: {err}")

    def terminate(self, instance_id: str) -> None:
        instance = self.details(instance_id)
        if not instance:
            return
        try:
            self.ec2_client.terminate_instances(InstanceIds=[instance_id])
            waiter = self.ec2_client.get_waiter('instance_terminated')
            waiter.wait(InstanceIds=[instance_id])
        except Exception as err:
            raise AWSDriverError(f"error terminating instance: {err}")

    def image_details(self, ami_id: str) -> dict:
        ami_filter = {
            'Name': 'image-id',
            'Values': [
                ami_id,
            ]
        }
        try:
            result = self.ec2_client.describe_images(Filters=[ami_filter], IncludeDeprecated=True)
        except Exception as err:
            raise AWSDriverError(f"error getting AMI details: {err}")

        return result['Images'][0]

    def get_password(self, instance_id: str, ssh_key: str) -> str:
        try:
            logger.info(f"Waiting for instance {instance_id} password")
            while True:
                result = self.ec2_client.get_password_data(InstanceId=instance_id)
                encrypted_password = result.get('PasswordData')
                if not encrypted_password:
                    time.sleep(5)
                    continue
                password_data = base64.b64decode(encrypted_password)
                return SSHUtil().decrypt_with_key(password_data, ssh_key)
        except Exception as err:
            raise AWSDriverError(f"error getting instance password: {err}")



##
##

import os
import logging
import uuid
import collections
import string
import random
import tarfile
import hashlib
import threading
import json
import importlib
from typing import Union, List, Callable
from uuid import UUID
from shutil import copyfile
from functools import wraps
from couchformation.provisioner.shell import RunShellCommand
from couchformation.exception import FatalError

logger = logging.getLogger('couchformation.util')
logger.addHandler(logging.NullHandler())
lock = threading.Lock()


class FileManagerError(FatalError):
    pass


def dict_merge(dict1, dict2):
    new_dict = dict(dict1)
    new_dict.update(dict2)
    return new_dict


def dict_merge_list(*dicts):
    res = collections.defaultdict(list)
    for d in dicts:
        for k, v in d.items():
            res[f"{k}_list"].append(v)
    return {k: set(v) for k, v in res.items()}


def dict_merge_not_none(source, target):
    new_dict = dict(target)
    for key, value in source.items():
        if new_dict.get(key) is None and source.get(key) is not None:
            new_dict[key] = value
    return new_dict


def parameter_to_dict(parameter: Union[str, None]) -> dict:
    if parameter is None or parameter == '':
        return {}
    result = {}
    for item in parameter.split(','):
        key, value = item.split(':')
        result[key] = value
    return result


def csv_dict_concat(parameter: dict, data: str) -> str:
    result = parameter_to_dict(data)
    result.update(parameter)
    return ','.join([f"{k}:{result[k]}" for k in result])


def progress_bar(iteration, total, decimals=1, length=100, fill='#', end="\r"):
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled = int(length * iteration // total)
    bar = fill * filled + '-' * (length - filled)
    print(f'\rProgress: |{bar}| {percent}% Complete', end=end)
    if iteration == total:
        print()


def dump_class_variables(data: dict):
    return json.dumps(data, indent=2, default=lambda o: '<not serializable>')


def synchronize() -> Callable:
    def lock_handler(func):
        @wraps(func)
        def f_wrapper(*args, **kwargs):
            with lock:
                return func(*args, **kwargs)
        return f_wrapper
    return lock_handler


class FileManager(object):

    def __init__(self):
        pass

    def make_dir(self, name: str):
        if not os.path.exists(name):
            path_dir = os.path.dirname(name)
            if not os.path.exists(path_dir):
                self.make_dir(path_dir)
            try:
                os.mkdir(name)
            except OSError:
                raise

    def dir_populate(self, path: str, command: Union[str, None] = None):
        self.make_dir(path)
        if command:
            cmd = command.split()
            RunShellCommand().cmd_exec(cmd, path)

    @staticmethod
    def copy_file(source: Union[str, bytes], destination: Union[str, bytes]) -> None:
        try:
            logger.debug(f"Copying {source} to {destination}")
            copyfile(source, destination)
        except Exception as err:
            raise FileManagerError(f"can not copy {source} to {destination}: {err}")

    @staticmethod
    def list_dir(dir_name: str):
        for name in os.listdir(dir_name):
            yield name

    @staticmethod
    def create_archive(file_name: str, root_dir: str, file_list: List[str]):
        path_list = []
        file_count = 0

        for path_name in file_list:
            if os.path.isdir(path_name):
                for root, dirs, files in os.walk(path_name):
                    for filename in dirs + files:
                        full_path = os.path.join(root, filename)
                        if not os.path.isfile(full_path):
                            continue
                        path_list.append(full_path)
                        file_count += 1
            else:
                path_list.append(path_name)
                file_count += 1

        progress_bar(0, file_count, length=50)
        with tarfile.open(file_name, mode='w:gz') as tar:
            for n, file_name in enumerate(path_list):
                rel_name = os.path.relpath(file_name, root_dir) if file_name.startswith(root_dir) else file_name
                tar.add(file_name, arcname=rel_name)
                progress_bar(n + 1, file_count, length=50)


class UUIDGen(object):

    def __init__(self):
        self._last_uuid = UUID('00000000000000000000000000000000')
        self._uuid = self._last_uuid

    def recompute(self, text):
        self._uuid = uuid.uuid5(self._last_uuid, text)
        self._last_uuid = self._uuid

    @property
    def uuid(self):
        return self._uuid

    @staticmethod
    def get_project_uid(project_name):
        return hashlib.md5(f"{uuid.getnode()}-{project_name}".encode()).hexdigest()[:10]

    @staticmethod
    def text_hash(text):
        return hashlib.md5(text.encode()).hexdigest()[:8]


class Synchronize(object):

    def __init__(self):
        pass

    def __enter__(self):
        lock.acquire()

    def __exit__(self, *args):
        lock.release()


class PasswordUtility(object):

    def __init__(self):
        pass

    @staticmethod
    def valid_password(password: str, min_length: int = 8, max_length: int = 64) -> bool:
        lower = 0
        upper = 0
        digit = 0
        special = 0
        if min_length <= len(password) <= max_length:
            for i in password:
                if i.islower():
                    lower += 1
                if i.isupper():
                    upper += 1
                if i.isdigit():
                    digit += 1
                if not i.isalnum():
                    special += 1

        if lower >= 1 and upper >= 1 and digit >= 1 and special >= 1:
            return True
        else:
            return False

    def generate(self, length: int = 8):
        while True:
            text = ''.join(random.choices(string.ascii_lowercase + string.ascii_uppercase + string.digits + '%@#', k=length))
            password = str(text)
            if self.valid_password(password, min_length=length):
                return password


class GenericAttrClass(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return dir(super(GenericAttrClass, self)) + list(self.keys())

    @property
    def as_dict(self):
        return {k: self[k] for k in self.keys()}


class CloudUtility(object):

    def __init__(self):
        pass

    @staticmethod
    def get_default_region(cloud: str) -> Union[str, None]:
        module_name = f"couchformation.{cloud.lower()}"
        try:
            module = importlib.import_module(module_name)
            if hasattr(module, "default_region"):
                return module.default_region
        except ModuleNotFoundError:
            pass
        return None



##
##

import logging
import re
import googleapiclient.discovery
import googleapiclient.errors
from datetime import datetime
from typing import List, Union
from couchformation.gcp.driver.base import CloudBase, GCPDriverError, EmptyResultSet
from couchformation.gcp.driver.constants import GCPImageProjects
import couchformation.constants as C

logger = logging.getLogger('couchformation.gcp.driver.image')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class Image(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, project: Union[str, None] = None, architecture: str = 'x86-64') -> List[dict]:
        image_list = []
        if not project:
            project = self.gcp_project

        request = self.gcp_client.images().list(project=project)

        while request is not None:
            try:
                response = request.execute()
            except Exception as err:
                raise GCPDriverError(f"error getting images: {err}")
            if response.get('items') is None:
                break
            for image in response['items']:
                if 'deprecated' in image:
                    if (image['deprecated']['state'] == "DEPRECATED") or (image['deprecated']['state'] == "OBSOLETE"):
                        continue
                _image_is_arm = re.search('arm64', image['name'])
                if architecture == 'arm64' and not _image_is_arm:
                    continue
                elif _image_is_arm:
                    continue
                image_block = {'name': image['name'],
                               'link': image['selfLink'],
                               'date': image['creationTimestamp']}
                image_block.update(self.process_labels(image))
                image_list.append(image_block)
            request = self.gcp_client.images().list_next(previous_request=request, previous_response=response)

        if len(image_list) == 0:
            raise EmptyResultSet(f"no images found")

        return image_list

    def details(self, image: str, project: Union[str, None] = None) -> dict:
        if not project:
            project = self.gcp_project

        try:
            request = self.gcp_client.images().get(project=project, image=image)
            image = request.execute()
        except Exception as err:
            if isinstance(err, googleapiclient.errors.HttpError):
                error_details = err.error_details[0].get('reason')
                if error_details == "notFound":
                    raise EmptyResultSet(f"image {image} not found")
            raise GCPDriverError(f"image detail error: {err}")

        image_block = {'name': image['name'],
                       'link': image['selfLink'],
                       'date': image['creationTimestamp']}
        image_block.update(self.process_labels(image))

        return image_block

    def delete(self, image: str) -> None:
        try:
            request = self.gcp_client.images().delete(project=self.gcp_project, image=image)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except Exception as err:
            raise GCPDriverError(f"error deleting image: {err}")

    def list_standard(self, architecture: str = 'x86_64', os_id: str = None, os_version: str = None):
        result_list = []
        for image_type in GCPImageProjects.projects:
            if os_id and image_type['os_id'] != os_id:
                continue
            image_list = self.list(project=image_type['project'], architecture=architecture)
            for version in C.OS_VERSION_LIST[image_type['os_id']]:
                if os_version and version != os_version:
                    continue
                filtered_images = []
                for image in image_list:
                    logger.debug(f"Found image -> {image['name']}")
                    match = re.search(image_type['pattern'], image['name'])
                    if match and (match.group(1) == version or match.group(1) == ''.join([letter for letter in '22.04' if letter.isalnum()])):
                        filtered_images.append(image)
                if len(filtered_images) > 0:
                    filtered_images.sort(key=lambda i: datetime.fromisoformat(i['date']))
                    result_image = filtered_images[-1]
                    result_image.update(dict(
                        os_id=image_type['os_id'],
                        os_version=version,
                        os_user=image_type['user'],
                        image_project=image_type['project']
                    ))
                    result_list.append(result_image)
                if len(result_list) > 0:
                    logger.debug(f"Selected image -> {result_list[-1]}")
                return result_list[-1]

    @staticmethod
    def image_user(os_id: str):
        return next((image_type['user'] for image_type in GCPImageProjects.projects if image_type['os_id'] == os_id), None)



##
##

import logging
import base64
import time
import googleapiclient.errors
import datetime
import copy
import json
from typing import Union
from couchformation.gcp.driver.base import CloudBase, GCPDriverError
from couchformation.ssh import SSHUtil

logger = logging.getLogger('couchformation.gcp.driver.instance')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class Instance(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def run(self,
            name: str,
            image_project: str,
            image_name: str,
            sa_email: Union[str, None],
            zone: str,
            vpc: str,
            subnet: str,
            username: str,
            ssh_key: str,
            swap_disk,
            data_disk,
            root_size="256",
            disk_type: str = "pd-ssd",
            machine_type="n2-standard-2",
            virtualization: bool = False):
        operation = {}
        instance_body = {
            "name": name,
            "zone": zone,
            "networkInterfaces": [
                {
                    "network": f"projects/{self.gcp_project}/global/networks/{vpc}",
                    "subnetwork": f"regions/{self.gcp_region}/subnetworks/{subnet}",
                    "accessConfigs": [
                        {
                            "name": "external-nat",
                            "type": "ONE_TO_ONE_NAT",
                            "networkTier": "PREMIUM"
                        }
                    ]
                }
            ],
            "metadata": {
                "items": [
                    {
                        "key": "ssh-keys",
                        "value": f"{username}:{ssh_key}"
                    }
                ]
            },
            "disks": [
                {
                    "boot": True,
                    "initializeParams": {
                        "sourceImage": f"projects/{image_project}/global/images/{image_name}",
                        "diskType": f"zones/{zone}/diskTypes/{disk_type}",
                        "diskSizeGb": str(round(float(root_size)))
                    },
                    "autoDelete": True
                },
                {
                    "source": f"zones/{zone}/disks/{swap_disk}"
                },
                {
                    "source": f"zones/{zone}/disks/{data_disk}"
                }
            ],
            "machineType": f"zones/{zone}/machineTypes/{machine_type}"
        }

        if virtualization:
            instance_body.update({
                "advancedMachineFeatures": {
                    "enableNestedVirtualization": True
                }
            })

        if sa_email:
            instance_body.update({
                "serviceAccounts": [
                    {
                        "email": sa_email,
                        "scopes": [
                            "https://www.googleapis.com/auth/cloud-platform"
                        ]
                    }
                ]
            })

        try:
            request = self.gcp_client.instances().insert(project=self.gcp_project, zone=zone, body=instance_body)
            operation = request.execute()
            self.wait_for_zone_operation(operation['name'], zone)
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create instance: {err}")
        except Exception as err:
            raise GCPDriverError(f"error creating instance: {err}")

        return operation.get('targetLink')

    def details(self, instance: str, zone: str) -> Union[dict, None]:
        try:
            request = self.gcp_client.instances().get(project=self.gcp_project, zone=zone, instance=instance)
            response = request.execute()
            return response
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find instance: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting instance details: {err}")

    def find(self, instance: str) -> Union[dict, None]:
        for zone in self.gcp_zone_list:
            result = self.details(instance, zone)
            if result:
                return result
        return None

    def terminate(self, instance: str, zone: str) -> None:
        try:
            request = self.gcp_client.instances().delete(project=self.gcp_project, zone=zone, instance=instance)
            operation = request.execute()
            self.wait_for_zone_operation(operation['name'], zone)
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not terminate instance: {err}")
        except Exception as err:
            raise GCPDriverError(f"error terminating instance: {err}")

    def gen_password(self, user: str, instance: str, zone: str, sa_email: str, ssh_key: str):
        instance_ref = self.details(instance, zone)
        old_metadata = instance_ref['metadata']

        mod, exp = SSHUtil().get_mod_exp(ssh_key)
        modulus = base64.b64encode(mod)
        exponent = base64.b64encode(exp)

        utc_now = datetime.datetime.now(datetime.timezone.utc)
        expire_time = utc_now + datetime.timedelta(minutes=5)
        expire = expire_time.strftime('%Y-%m-%dT%H:%M:%SZ')

        metadata_entry = {'userName': user,
                          'modulus': modulus.decode('utf-8'),
                          'exponent': exponent.decode('utf-8'),
                          'email': sa_email,
                          'expireOn': expire}

        new_metadata = copy.deepcopy(old_metadata)
        new_metadata['items'] = [{
            'key': "windows-keys",
            'value': json.dumps(metadata_entry)
        }]

        request = self.gcp_client.instances().setMetadata(project=self.gcp_project,
                                                          zone=zone,
                                                          instance=instance,
                                                          body=new_metadata)
        operation = request.execute()
        self.wait_for_zone_operation(operation['name'], zone)

        logger.info(f"Waiting for instance {instance} password")
        while True:
            request = self.gcp_client.instances().getSerialPortOutput(project=self.gcp_project,
                                                                      zone=zone,
                                                                      instance=instance,
                                                                      port=4)
            operation = request.execute()
            serial_port_output = operation['contents']
            if len(serial_port_output) != 0:
                break
            else:
                time.sleep(2)

        output = serial_port_output.split('\n')
        for data in output:
            try:
                entry = json.loads(data)
                if modulus.decode('utf-8') == entry['modulus']:
                    enc_password = entry['encryptedPassword']
                    decoded_password = base64.b64decode(enc_password)
                    password = SSHUtil().decrypt_with_rsa(decoded_password, ssh_key)
                    return password.decode('utf-8')
            except ValueError:
                pass



##
##

import logging

logger = logging.getLogger('couchformation.executor.worker')
logger.addHandler(logging.NullHandler())


def main(module, instance, method, *args, **kwargs):
    m = __import__(module, fromlist=[""])
    i = getattr(m, instance)
    obj = i(*args, **kwargs)
    f = getattr(obj, method)
    return f()


def get_class(module, instance, *args, **kwargs):
    m = __import__(module, fromlist=[""])
    i = getattr(m, instance)
    obj = i(*args, **kwargs)
    return obj


def run_method(obj, method, *args, **kwargs):
    f = getattr(obj, method)
    return f(*args, **kwargs)



##
##

import io
import os
import getpass
import stat
import sys
import logging
import platform
import tarfile
import time
import json
import csv
import plistlib
from typing import List, Any, Tuple
from couchformation.config import get_log_dir, get_root_dir, get_base_dir
from pathlib import Path
from datetime import datetime
import couchformation.kvdb as kvdb
if os.name == 'nt':
    group_name = "staff"
else:
    import grp
    group_name = grp.getgrgid(os.getgid())[0]

logger = logging.getLogger('couchformation.support')
logger.addHandler(logging.NullHandler())


class CreateDebugPackage(object):

    def __init__(self):
        now = datetime.now()
        time_string = now.strftime("%m%d%y%H%M%S")
        self.os_release = "/etc/os-release"
        self.macos_release = "/System/Library/CoreServices/SystemVersion.plist"
        self.log_dir = get_log_dir()
        self.root_dir = get_root_dir()
        self.state_dir = get_base_dir()
        self.debug_file = os.path.join(self.log_dir, "formation.log")
        self.crash_file = os.path.join(self.log_dir, "crash.log")
        self.tgz_file = f"cf-debug-{time_string}.tar.gz"
        self.support_bundle = os.path.join(Path.home(), self.tgz_file)
        self.file_list = []

        logger.debug(f"Python version: {''.join(sys.version.splitlines())}")
        logger.debug(f"System: {platform.system()} Release: {platform.release()}")

        if os.path.exists(self.debug_file):
            self.file_list.append(('path', self.debug_file))
        if os.path.exists(self.crash_file):
            self.file_list.append(('path', self.crash_file))

        state_dir_listing = self.state_file_list()
        self.file_list.append(("state_dir_listing.log", state_dir_listing))

        state_data_dump = self.state_data_dump()
        self.file_list.append(("state_data.json", state_data_dump))

        os_info = self.os_info()
        self.file_list.append(("os_info.txt", os_info))

    def create_snapshot(self):
        logger.info(f"Writing support bundle to {self.support_bundle}")
        self.support_archive(self.support_bundle, self.root_dir, self.file_list)

    def os_info(self):
        data = io.BytesIO()

        if os.path.exists(self.os_release):
            with open(self.os_release) as f:
                reader = csv.reader(f, delimiter="=")
                for rows in reader:
                    if len(rows) < 2:
                        continue
                    data.write(f"{rows[0]}: {rows[1]}\n".encode())
        elif os.path.exists(self.macos_release):
            with open(self.macos_release, 'rb') as f:
                sys_info = plistlib.load(f)
            data.write(f"NAME      : {os.uname().sysname}\n".encode())
            data.write(f"VERSION   : {os.uname().version}\n".encode())
            data.write(f"ID        : {sys_info.get('ProductName')}\n".encode())
            data.write(f"VERSION_ID: {sys_info.get('ProductVersion')}\n".encode())

        uname = platform.uname()
        data.write(f"Node     : {uname.node}\n".encode())
        data.write(f"System   : {uname.system}\n".encode())
        data.write(f"Version  : {uname.version}\n".encode())
        data.write(f"Release  : {uname.release}\n".encode())
        data.write(f"Machine  : {uname.machine}\n".encode())
        data.write(f"Processor: {uname.processor}\n".encode())

        data.seek(0)
        return data

    def state_file_list(self):
        data = io.BytesIO()

        for root, dirs, files in os.walk(self.state_dir):
            for filename in dirs + files:
                full_path = os.path.join(root, filename)
                try:
                    st = os.stat(full_path)
                    is_dir = 'd' if stat.S_ISDIR(st.st_mode) else '-'
                    d = {'7': 'rwx', '6': 'rw-', '5': 'r-x', '4': 'r--', '0': '---'}
                    perm = str(oct(st.st_mode)[-3:])
                    prefix = is_dir + ''.join(d.get(x, x) for x in perm)
                    user = getpass.getuser()
                    group = group_name
                    data.write(f"{prefix} {user:<9} {group:<9} {full_path}\n".encode())
                except Exception as err:
                    data.write(f"{full_path}: {err}\n".encode())

        data.seek(0)
        return data

    def state_data_dump(self):
        data = io.BytesIO()
        obj_list = []

        for root, dirs, files in os.walk(self.state_dir):
            for filename in dirs + files:
                full_path = os.path.join(root, filename)
                st = os.stat(full_path)
                if stat.S_ISREG(st.st_mode):
                    with open(full_path, 'rb') as data_file:
                        header = data_file.read(16)
                        if header.decode('utf-8').startswith('SQLite format 3'):
                            try:
                                block = {
                                    'file_name': full_path,
                                    'documents': {}
                                }
                                for doc in kvdb.documents(full_path):
                                    block['documents'][doc.document_id] = {}
                                    null_c = 1
                                    for key, value in doc.items():
                                        if not key:
                                            key = f"null_{null_c}"
                                            null_c += 1
                                        if key == 'host_password' or key == 'password':
                                            continue
                                        block['documents'][doc.document_id].update({key: value})
                                obj_list.append(block)
                            except Exception as err:
                                block = {
                                    'file_name': full_path,
                                    'error': err
                                }
                                obj_list.append(block)

        data.write(json.dumps(obj_list, indent=2).encode())
        data.seek(0)
        return data

    @staticmethod
    def support_archive(file_name: str, root_dir: str, file_list: List[Tuple[str, Any]]):
        with tarfile.open(file_name, mode='w:gz') as tar:
            for n, file_tuple in enumerate(file_list):
                file_text = file_tuple[0]
                if isinstance(file_tuple[1], io.BytesIO):
                    file_data: io.BytesIO = file_tuple[1]
                    tar_info = tarfile.TarInfo(name=f"log/{file_text}")
                    tar_info.mtime = time.time()
                    tar_info.uid = 0
                    tar_info.gid = 0
                    tar_info.size = file_data.getbuffer().nbytes
                    tar.addfile(tarinfo=tar_info, fileobj=file_data)
                else:
                    file_path = file_tuple[1]
                    rel_name = os.path.relpath(file_path, root_dir) if file_path.startswith(root_dir) else file_path
                    tar.add(file_path, arcname=rel_name)



##
##

import logging
import botocore.exceptions
from typing import Union, List
from botocore.exceptions import ClientError
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSTagStruct, AWSTag

logger = logging.getLogger('couchformation.aws.driver.sshkey')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class SSHKey(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def list(self, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        key_list = []

        try:
            key_pairs = self.ec2_client.describe_key_pairs()
        except Exception as err:
            raise AWSDriverError(f"error getting key pairs: {err}")

        for key in key_pairs['KeyPairs']:
            key_block = {'name': key['KeyName'],
                         'id': key['KeyPairId'],
                         'fingerprint': key['KeyFingerprint'],
                         'pubkey': key.get('PublicKey')}
            if filter_keys_exist:
                if not all(key in key_block for key in filter_keys_exist):
                    continue
            key_list.append(key_block)

        if len(key_list) == 0:
            raise EmptyResultSet(f"no SSH keys found")

        return key_list

    def create(self, name: str, ssh_key: str, tags: Union[dict, None] = None) -> str:
        key_block = {}
        tag_build = AWSTagStruct.build("key-pair")
        tag_build.add(AWSTag("Name", name))

        if tags:
            for key, value in tags.items():
                tag_build.add(AWSTag(key, value))

        key_tag = [tag_build.as_dict]

        try:
            result = self.ec2_client.import_key_pair(KeyName=name,
                                                     PublicKeyMaterial=ssh_key.encode('utf-8'),
                                                     TagSpecifications=key_tag)
            key_block = {'name': result['KeyName'],
                         'id': result['KeyPairId'],
                         'fingerprint': result['KeyFingerprint']}
        except ClientError as err:
            if err.response['Error']['Code'] == 'InvalidKeyPair.Duplicate':
                key_block = self.details(name)
            else:
                raise
        except Exception as err:
            AWSDriverError(f"error importing key pair: {err}")

        return key_block['name']

    def create_native(self, name: str) -> dict:
        key_block = {}
        try:
            result = self.ec2_client.create_key_pair(KeyName=name)
            key_block = {'name': result['KeyName'],
                         'id': result['KeyPairId'],
                         'fingerprint': result['KeyFingerprint'],
                         'key': result['KeyMaterial']}
        except Exception as err:
            AWSDriverError(f"error creating key pair: {err}")

        return key_block

    def details(self, key_name: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_key_pairs(KeyNames=[key_name])
            key_result = result['KeyPairs'][0]
            key_block = {'name': key_result['KeyName'],
                         'id': key_result['KeyPairId'],
                         'fingerprint': key_result['KeyFingerprint']}
            return key_block
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error deleting key pair: {err}")

    def delete(self, name: str) -> None:
        key = self.details(name)
        if not key:
            return
        try:
            self.ec2_client.delete_key_pair(KeyName=name)
        except Exception as err:
            raise AWSDriverError(f"error deleting key pair: {err}")

    def instances_by_key(self, name: str):
        instances = []
        extra_args = {}
        key_filter = [
            {
                'Name': 'key-name',
                'Values': [
                    name,
                ]
            },
            {
                'Name': 'instance-state-name',
                'Values': [
                    'running',
                    'pending'
                ]
            }
        ]

        try:
            while True:
                result = self.ec2_client.describe_instances(**extra_args, Filters=key_filter)
                for reservation in result.get('Reservations', []):
                    instances.extend(reservation.get('Instances', []))
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting instance list: {err}")

        return instances



##
##

import time
import asyncio
import logging
from typing import Callable
from functools import wraps

logger = logging.getLogger('couchformation.retry')
logger.addHandler(logging.NullHandler())


def retry_inline(func, *args, retry_count=10, factor=0.01, **kwargs):
    for retry_number in range(retry_count + 1):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            if retry_number == retry_count:
                logger.debug(f"{func.__name__} retry limit exceeded: {err}")
                raise
            logger.debug(f"{func.__name__} will retry, number {retry_number + 1}")
            wait = factor
            wait *= (2 ** (retry_number + 1))
            time.sleep(wait)


def retry(retry_count=10,
          factor=0.01,
          allow_list=None,
          always_raise_list=None
          ) -> Callable:

    def retry_handler(func):
        if not asyncio.iscoroutinefunction(func):
            @wraps(func)
            def f_wrapper(*args, **kwargs):
                for retry_number in range(retry_count + 1):
                    try:
                        return func(*args, **kwargs)
                    except Exception as err:
                        if always_raise_list and isinstance(err, always_raise_list):
                            raise

                        if allow_list and not isinstance(err, allow_list):
                            raise

                        if retry_number == retry_count:
                            logger.debug(f"{func.__name__} retry limit exceeded")
                            raise

                        logger.debug(f"{func.__name__} will retry, number {retry_number + 1}")
                        wait = factor
                        wait *= (2 ** (retry_number + 1))
                        time.sleep(wait)

            return f_wrapper
        else:
            @wraps(func)
            async def f_wrapper(*args, **kwargs):
                for retry_number in range(retry_count + 1):
                    try:
                        return await func(*args, **kwargs)
                    except Exception as err:
                        if always_raise_list and isinstance(err, always_raise_list):
                            raise

                        if allow_list and not isinstance(err, allow_list):
                            raise

                        if retry_number == retry_count:
                            logger.debug(f"{func.__name__} retry limit exceeded")
                            raise

                        logger.debug(f"{func.__name__} will retry, number {retry_number + 1}")
                        wait = factor
                        wait *= (2 ** (retry_number + 1))
                        time.sleep(wait)

            return f_wrapper
    return retry_handler



##
##

import logging
from typing import List, Union
import googleapiclient.errors
from couchformation.gcp.driver.base import CloudBase, GCPDriverError, EmptyResultSet

logger = logging.getLogger('couchformation.gcp.driver.network')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class Network(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self) -> List[dict]:
        network_list = []

        try:
            request = self.gcp_client.networks().list(project=self.gcp_project)
            while request is not None:
                response = request.execute()

                for network in response['items']:
                    subnet_list = []
                    for subnet in network.get('subnetworks', []):
                        subnet_name = subnet.rsplit('/', 4)[-1]
                        region_name = subnet.rsplit('/', 4)[-3]
                        if region_name != self.region:
                            continue
                        result = Subnet(self.parameters).details(subnet_name)
                        subnet_list.append(result)
                    network_block = {'cidr': network.get('IPv4Range', None),
                                     'name': network['name'],
                                     'description': network.get('description', None),
                                     'subnets': subnet_list,
                                     'id': network['id']}
                    network_list.append(network_block)
                request = self.gcp_client.networks().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverError(f"error listing networks: {err}")

        if len(network_list) == 0:
            raise EmptyResultSet(f"no networks found")
        else:
            return network_list

    @property
    def cidr_list(self):
        try:
            for network in self.list():
                for item in Subnet(self.parameters).list(network['name']):
                    yield item['cidr']
        except EmptyResultSet:
            return iter(())

    def create(self, name: str) -> str:
        operation = {}
        network_body = {
            "name": name,
            "autoCreateSubnetworks": False
        }
        try:
            request = self.gcp_client.networks().insert(project=self.gcp_project, body=network_body)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
            return operation.get('targetLink')
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create network: {err}")
        except Exception as err:
            raise GCPDriverError(f"error creating network: {err}")

        return operation.get('targetLink')

    def delete(self, network: str) -> None:
        try:
            request = self.gcp_client.networks().delete(project=self.gcp_project, network=network)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not delete network: {err}")
        except Exception as err:
            raise GCPDriverError(f"error deleting network: {err}")

    def details(self, network: str) -> Union[dict, None]:
        try:
            request = self.gcp_client.networks().get(project=self.gcp_project, network=network)
            result = request.execute()
            return result
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find network: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting network: {err}")

    def add_peering(self, name: str, network: str, peer_project: str, peer_network: str) -> None:
        peering_body = {
            "networkPeering": {
                "name": name,
                "network": f"projects/{peer_project}/global/networks/{peer_network}",
                "exchangeSubnetRoutes": True
            }
        }
        try:
            request = self.gcp_client.networks().addPeering(project=self.gcp_project, network=network, body=peering_body)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not add network peering: {err}")
        except Exception as err:
            raise GCPDriverError(f"error adding network peering: {err}")

    def remove_peering(self, name: str, network: str) -> None:
        remove_body = {
            "name": name
        }
        try:
            request = self.gcp_client.networks().removePeering(project=self.gcp_project, network=network, body=remove_body)
            operation = request.execute()
            self.wait_for_global_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not remove network peering: {err}")
        except Exception as err:
            raise GCPDriverError(f"error removing network peering: {err}")


class Subnet(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, network: str, region: Union[str, None] = None) -> List[dict]:
        subnet_list = []

        try:
            request = self.gcp_client.subnetworks().list(project=self.gcp_project, region=self.gcp_region)
            while request is not None:
                response = request.execute()
                for subnet in response['items']:
                    network_name = subnet['network'].rsplit('/', 1)[-1]
                    region_name = subnet['region'].rsplit('/', 1)[-1]
                    if region:
                        if region != region_name:
                            continue
                    if network != network_name:
                        continue
                    subnet_block = {'cidr': subnet['ipCidrRange'],
                                    'name': subnet['name'],
                                    'description': subnet.get('description', None),
                                    'gateway': subnet['gatewayAddress'],
                                    'network': network_name,
                                    'region': region_name,
                                    'id': subnet['id']}
                    subnet_list.append(subnet_block)
                request = self.gcp_client.subnetworks().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverError(f"error listing subnets: {err}")

        if len(subnet_list) == 0:
            raise EmptyResultSet(f"no subnets found")
        else:
            return subnet_list

    def create(self, name: str, network: str, cidr: str) -> str:
        operation = {}
        network_info = Network(self.parameters).details(network)
        subnetwork_body = {
            "name": name,
            "network": network_info['selfLink'],
            "ipCidrRange": cidr,
            "region": self.gcp_region
        }
        try:
            request = self.gcp_client.subnetworks().insert(project=self.gcp_project, region=self.gcp_region, body=subnetwork_body)
            operation = request.execute()
            self.wait_for_regional_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create subnet: {err}")
        except Exception as err:
            raise GCPDriverError(f"error creating subnet: {err}")

        return operation.get('targetLink')

    def delete(self, subnet: str) -> None:
        try:
            request = self.gcp_client.subnetworks().delete(project=self.gcp_project, region=self.gcp_region, subnetwork=subnet)
            operation = request.execute()
            self.wait_for_regional_operation(operation['name'])
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not delete subnet: {err}")
        except Exception as err:
            raise GCPDriverError(f"error deleting subnet: {err}")

    def details(self, subnet: str) -> Union[dict, None]:
        try:
            request = self.gcp_client.subnetworks().get(project=self.gcp_project, region=self.gcp_region, subnetwork=subnet)
            result = request.execute()
            network_name = result['network'].rsplit('/', 1)[-1]
            region_name = result['region'].rsplit('/', 1)[-1]
            subnet_block = {'cidr': result['ipCidrRange'],
                            'name': result['name'],
                            'description': result.get('description', None),
                            'gateway': result['gatewayAddress'],
                            'network': network_name,
                            'region': region_name,
                            'id': result['id']}
            return subnet_block
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find subnet: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting subnet: {err}")
